"""\
Utility functions to work with "wireshark" network capture files

Initial Version: Costas Skarakis 23/2/2020
"""
import os.path
import platform
import re
import subprocess as sb
import json
from difflib import Differ
from pprint import pprint

from common.util import XmlBody
from sip.SipMessage import SipMessage
from sip.SipParser import buildMessage

replacement_set = {"diff": {r"Call-ID: .*": "Call-ID: {callId}",
                            r"sip:.*@[\w\d\.]+:\d+": r"sip:{user}@{dest_ip}:{dest_port}",
                            r"sip:.*@[\w\d\.]+": r"sip:{user}@{dest_ip}",
                            r": .* <": ": \"{user}\" <",
                            r"transport=\w{3}": "transport={transport}",
                            r"SIP/2\.0/\w{3}": "SIP/2.0/{transport}",
                            r"(From.*)tag=.*": r"\1tag={fromTag}",
                            r"(To.*)tag=.*": r"\1tag={toTag}",
                            r"(Via.*)branch=.*": r"\1branch={viaBranch}",
                            r"CSeq: \d+ ([A-Z]+)": r"CSeq: {cseq} \1",
                            r"Max-Forwards: \d+": r"Max-Forwards: {max_forwards}",
                            r"\d{1,3}\.\d{1,3}\.\d{1,3}\.\d{1,3}:\d+": r"{ip}:{port}",
                            r"\d{1,3}\.\d{1,3}\.\d{1,3}\.\d{1,3}": r"{ip}",
                            r"^([\w-]+): \d+": r"\1: {\1}",
                            r"\d{5,}": "{num}",  # {five_plus_digit_number
                            r"[0-9A-Fa-f]{10,}": "{hash}",  # 10_plus_character_hex_string
                            r"[\d\.-_/\\~]{3,}": "{hash}"
                            # 3_plus_character_string_with_only_digits_and_special_characters
                            },
                   "testcase": {r"Call-ID: .*": "Call-ID: {callId}",
                                r"(REGISTER|NOTIFY|SUBSCRIBE)sip:.*@[\w\d\.]+:\d+":
                                    r"\1 sip:{user}@{dest_ip}:{dest_port}",
                                r"(REGISTER|NOTIFY|SUBSCRIBE) sip:.*@[\w\d\.]+":
                                    r"\1 sip:{user}@{dest_ip}",
                                # All request but NOTIFY SUBSCRIBE and REGISTER have the B side user
                                r"([A-Z]+) sip:.*@[\w\d\.]+:\d+": r"\1 sip:{userB}@{dest_ip}:{dest_port}",
                                r"([A-Z]+) sip:.*@[\w\d\.]+": r"\1 sip:{userB}@{dest_ip}",
                                # To header includes the B side user
                                r"To: \".*\" <sip:.*@[\w\d\.]+:\d+":
                                    r"To: \"{userB}\" <sip:{userB}@{dest_ip}:{dest_port}",
                                r"To: <sip:.*@[\w\d\.]+":
                                    r"To: <sip:{userB}@{dest_ip}",
                                r"sip:.*@[\w\d\.]+:\d+": r"sip:{user}@{dest_ip}:{dest_port}",
                                r"sip:.*@[\w\d\.]+": r"sip:{user}@{dest_ip}",
                                r": .* <": ": \"{user}\" <",
                                r"transport=\w{3}": "transport={transport}",
                                r"SIP/2\.0/\w{3}": "SIP/2.0/{transport}",
                                r"(From.*)tag=.*": r"\1tag={fromTag}",
                                r"(To.*)tag=.*": r"\1tag={toTag}",
                                r"(Via.*)branch=.*": r"\1branch={viaBranch}",
                                r"CSeq: \d+ ([A-Z]+)": r"CSeq: {cseq} \1",
                                r"Max-Forwards: \d+": r"Max-Forwards: {max_forwards}",
                                r"\d{1,3}\.\d{1,3}\.\d{1,3}\.\d{1,3}:\d+": r"{ip}:{port}",
                                r"\d{1,3}\.\d{1,3}\.\d{1,3}\.\d{1,3}": r"{ip}",
                                r"^([\w-]+): \d+": r"\1: {\1}",
                                r"\d{5,}": "{num}",  # {five_plus_digit_number
                                r"[0-9A-Fa-f]{10,}": "{hash}",  # 10_plus_character_hex_string
                                }
                   }


def open_cap_file(filename, tshark_path, tshark_filter=None):
    outfile = filename.rsplit(".", 1)[0] + ".json"
    if not tshark_path:
        print("tshark_path not provided. Program installation path must exist in PATH environment variable")
        tshark = "tshark"
    else:
        tshark = os.path.join(tshark_path, "tshark")
    if " " in tshark:
        tshark = '"' + tshark + '"'
    if tshark_filter:
        filter = "-Y \"{}\"".format(tshark_filter)
    else:
        filter = ""
    cmd = "{} -r {} {}  -T json > {}".format(tshark, filename, filter, outfile)
    print(cmd)
    out, err = sb.Popen(cmd, shell=True, stdout=sb.PIPE, stderr=sb.PIPE).communicate()
    if out:
        print(out)
    if err:
        print(err)
    return os.path.join(".", outfile)


def get_from_cap_file(filename, wireshark_filter=None, tshark_path=""):
    """
    Read a tethereal/tshark/wireshark file and return all the messages as a list of strings
    Requires tshark installed.

    :param filename:  the name of file (pcapng content)
    :param tshark_path: the location of tshark.
    :param wireshark_filter: apply a wireshark filter to the output
    :return: a list of messages as a list of strings
    """
    j_output = open_cap_file(filename, tshark_path, wireshark_filter)
    return get_sip_from_json_file(j_output)


def get_sip_from_json_file(filename):
    """
    Read a tethereal/tshark/wireshark file and return all the messages as a list of strings
    Requires tshark installed.

    :param filename:  the name of file (json content)
    :return: a list of messages as a list of strings
    """
    with open(filename, "r") as j_file:
        j_obj = json.load(j_file)
    result = []
    for j_msg in j_obj:
        try:
            this_msg = assemble_message_from_json(j_msg, appl="sip")
            result.append(this_msg)
        except KeyError:
            continue
    return result


def assemble_message_from_json(j_msg, appl):
    if appl == "sip":
        this_msg = ""
        header_d = j_msg["_source"]["layers"]["sip"]
        if isinstance(header_d, dict):
            if "sip.Request-Line" in header_d.keys():
                this_msg += header_d["sip.Request-Line"]

            elif "sip.Status-Line" in header_d.keys():
                this_msg += header_d["sip.Status-Line"]

            this_msg += '\r\n'
            # this seems to include the body as well, at least in trace with notify messages with xml bodies
            this_msg += header_d["sip.msg_hdr"]
        else:
            print("Error parsing sip message with contents:")
            print("\n{}\n".format(header_d))

    elif appl == "http":
        for key in j_msg["_source"]["layers"]["http"]:
            if "HTTP/" in key:
                request_uri = key
        this_msg = request_uri

        if "http.request.line" in j_msg["_source"]["layers"]["http"][request_uri]:
            for line in j_msg["_source"]["layers"]["http"][request_uri]["http.request.line"]:
                this_msg += line
        else:
            for line in j_msg["_source"]["layers"]["http"][request_uri]["http.response.line"]:
                this_msg += line
        this_msg += '\r\n'
        # this seems to include the body as well, at least in trace with notify messages with xml bodies
        this_msg += j_msg["_source"]["layers"]["http"][request_uri]["http.file_data"]

    elif appl == "rtp":
        this_msg = "RTP payload"
    else:
        this_msg = appl + " content"

    return this_msg


def make_sip_message_template(sip_message, purpose="diff"):
    """
    Takes a SipMessage instance and returns a string with all dynamic elements removed and replaced with placeholders
    :param sip_message: The SipMessage instance or a sip message as string
    :return: A string template
    """
    if isinstance(sip_message, str):
        contents = sip_message
    elif isinstance(sip_message, SipMessage):
        contents = sip_message.contents()
    else:
        print("Invalid sip message type", type(sip_message))
        return -1
    replacements = replacement_set[purpose]
    result = ""
    for line in contents.split("\r\n"):
        for item in replacements:
            patrn, rplmnt = item, replacements[item]
            line = re.sub(patrn, rplmnt, line, re.IGNORECASE)
        result += line + "\r\n"
    return result + "\r\n"


def get_msg_list_from_file(trace_file, input_format, wireshark_filter=None, tshark_path=""):
    if not tshark_path and platform.system() == "Windows":
        tshark_path = r"C:\Program Files\Wireshark"
    if input_format == "pcapng":
        list1 = get_from_cap_file(filename=trace_file, tshark_path=tshark_path, wireshark_filter=wireshark_filter)
    elif input_format == "json":
        list1 = get_sip_from_json_file(filename=trace_file)
    else:
        print("Invalid input file type: {}. Supported formats are pcapng and json".format(input_format))
        return None
    return list1


def diff(ref_trace, check_trace, input_format="pcapng", tshark_path="", filters={}):
    """
    Diff two pcapng or json files containing network captured data

    :param ref_trace: The first file. Considered as reference.
    :param check_trace: The second file. Checked to contain the same flow as the first file.
    :param input_format: Type of files to be checked. Can be "pcapng"  (default) or "json"
    :param tshark_path: Needed in case the input type is "pcapng" so that the files can be converted to json format
    :param filters: A list of filters to remove noise from messages.
    :return: The result of a list of checks
    """
    list1 = iter(get_msg_list_from_file(trace_file=ref_trace, tshark_path=tshark_path, input_format=input_format))
    list2 = iter(get_msg_list_from_file(trace_file=check_trace, tshark_path=tshark_path, input_format=input_format))
    if list1 is None or list2 is None:
        return None
    count = 0
    if "Call-ID" not in filters:
        filters["Call-ID"] = []
    for msg in list2:
        sip_msg = buildMessage(msg, {})
        if msg_filter(sip_msg, filters) is None:
            if sip_msg.type == "Request":
                filters["Call-ID"].append(sip_msg["Call-ID"])
            continue
        ref_msg = next(list1)
        ref_sip_msg = buildMessage(ref_msg, {})
        while msg_filter(ref_sip_msg, filters) is None:
            if ref_sip_msg.type == "Request":
                filters["Call-ID"].append(ref_sip_msg["Call-ID"])
            ref_msg = next(list1)
            ref_sip_msg = buildMessage(ref_msg, {})

        ref_msg_template = make_sip_message_template(ref_msg)
        msg_template = make_sip_message_template(msg)
        if msg != ref_msg:
            print("{:-^60}".format(" Difference found in message #{} ".format(count)))
            print("{:#^60}".format(" Reference message "))
            print(ref_msg)
            print("{:#^60}".format(" Matched message "))
            print(msg)

            print("{:#^60}".format(" Diff analysis "))
            d = Differ()
            for line in d.compare(ref_msg_template.splitlines(keepends=True),
                                  msg_template.splitlines(keepends=True)):
                print(line.strip())
            return False
        count += 1
    return True


def msg_filter(message, filter_data):
    """
    Used to filter a list of message based on given criteria

    :param message: A message (currently supported only sip)
    :param filter_data: A dictionary of lists with the filter data.
                        We can filter based on type of message, Call-ID header, or text found anywhere in the message.
        filter_data example:
        {"Message": ["NOTIFY", "100 Trying", "180"],
        "Call-ID": ["232ttwegwgweweggweg", "gqqh1h542h245h2", "2325235UIGBIU2390Hh45"],
        "Text": ["10.2.10.3", "302118814445", "Line10000", "keyset-info"]}
    :return: None if message matches the criteria, the message itself if it doesn't match
    """
    for excl_type in filter_data:
        if excl_type == "Message":
            for msg_type in filter_data["Message"]:
                if msg_type in message.get_status_or_method():
                    return None
        elif excl_type == "Text":
            for text in filter_data["Text"]:
                if text in message.contents():
                    return None
        elif excl_type == "Call-ID":
            for callID in filter_data["Call-ID"]:
                if message["Call-ID"] == callID:
                    return None
    return message


def check_in_trace(*conditions_list, check_trace, input_format="pcapng", tshark_path=""):
    """
    Look in the given trace file for a Message that matches the given conditions
    :param conditions_list: A list of dictionaries containing the conditions that a message in the trace should match.
                        Best described by an example:
           {"Message": "INVITE",                <- Message Req URI/Status Line should contain this string (re supported)
            "Headers": {"CSeq": "INVITE",       <- Message Header CSeq must contain this string (re)
                        "Supported": "timer"},  <- Message Header Supported must contain this string (re)
            "sdp": {"any": "PMCA",              <- Message sdp body must contain this text anywhere
                    "a_line": "PMCU",           <- Message sdp body must contain this text in an (a) line (re)
                    "o_line": "IP4"}            <- Message sdp body must contain this text in an (o) line
           "xml":  {"any": "ns:recording",      <- Message xml body must contain this text anywhere
                    "label tag": "audio",       <- Message xml body must contain this text in an a "label" tag
                    "label aor attr": "911@"}   <- Message xml body must contain this text in an "aor" attribute of
                                                    a "label" tag
            }
    :param check_trace: The trace containing the network capture under test
    :param input_format: Type of files to be checked. Can be "pcapng"  (default) or "json"
    :param tshark_path: Needed in case the input type is "pcapng" so that the files can be converted to json format
    :return:
    """
    result = []
    if input_format == "list":
        msg_list = check_trace
    else:
        msg_list = get_msg_list_from_file(check_trace, input_format=input_format, tshark_path=tshark_path)
    if not conditions_list:
        return msg_list
    for conditions in conditions_list:
        for msg_str in msg_list:
            try:
                if isinstance(msg_str, SipMessage):
                    msg = msg_str
                else:
                    msg = buildMessage(msg_str)
            except:
                print("Failed to parse", msg_str)
                raise
            match = True
            for check in conditions:
                if check == "Message":
                    if not re.search(conditions["Message"], msg.get_status_or_method(), re.IGNORECASE):
                        match = False
                        continue
                elif check == "Headers":
                    for header in conditions["Headers"]:
                        if header not in msg.headers:
                            match = False
                            continue
                        elif not re.search(conditions["Headers"][header], msg[header], re.IGNORECASE):
                            match = False
                            continue
                elif check == "sdp":
                    for sdpline in conditions["sdp"]:
                        if sdpline == "any":
                            if not re.search(conditions["sdp"]["any"], msg.body, re.IGNORECASE):
                                match = False
                                continue
                        elif sdpline.endswith("_line"):
                            line_type = sdpline[0]
                            if not re.search(line_type + "=.*" + conditions["sdp"][sdpline], msg.body, re.IGNORECASE):
                                match = False
                                continue

                elif check == "xml":
                    if "<?xml" not in msg.body:
                        match = False
                        continue
                    xml_part_of_body = "<?xml" + msg.body.split("<?xml")[1].split("\r\n\r\n")[0]
                    xml_body_obj = XmlBody(xml_part_of_body)
                    for xmlelement in conditions["xml"]:
                        if xmlelement == "any":
                            if not re.search(conditions["xml"]["any"], msg.body, re.IGNORECASE):
                                match = False
                                continue
                        elif xmlelement.endswith(" tag"):
                            xml_tag = xmlelement.split(" ")[0]
                            all_tags = xml_body_obj.get_all(xml_tag)
                            if not all_tags:
                                match = False
                                continue
                            elif not any(re.search(conditions["xml"][xmlelement], tag.text, re.IGNORECASE)
                                         for tag in all_tags):
                                match = False
                                continue
                        elif xmlelement.endswith(" attr"):
                            xml_tag = xmlelement.split(" ")[0]
                            xml_attr = xmlelement.split(" ")[1]
                            all_tags = xml_body_obj.get_all(xml_tag)
                            if not all_tags:
                                match = False
                                continue
                            elif not any(re.search(conditions["xml"][xmlelement], tag.get(xml_attr), re.IGNORECASE)
                                         for tag in all_tags
                                         if tag.get(xml_attr) is not None):
                                match = False
                                continue
                else:
                    print("Unsupported check", check)
                    raise

            if match:
                result.append(msg)
                break
    return result


def summarize_trace(filename, *tests, applications=("sip", "http", "rtp"), input_format="pcapng", tshark_path=None,
                    tshark_filter=None, delete_json=True):
    if not tshark_path and platform.system() == "Windows":
        tshark_path = r"C:\Program Files\Wireshark"
    if input_format == "json":
        with open(filename, "r") as j_file:
            j_obj = json.load(j_file)
    elif input_format == "pcapng":
        temp_file = open_cap_file(filename, tshark_path, tshark_filter)
        with open(temp_file, "r") as j_file:
            j_obj = json.load(j_file)
        if delete_json:
            os.remove(temp_file)
    else:
        print("Invalid input file type: {}. Supported formats are pcapng and json".format(input_format))
        return None
    result = {}
    for application in applications:
        result[application] = {"count": 0}
    for j_msg in j_obj:
        frame_protocols = j_msg["_source"]["layers"]["frame"]["frame.protocols"]
        if not any([application in frame_protocols for application in applications]):
            continue
        ip_layer = frame_protocols.split(":")[2]
        transport_layer = frame_protocols.split(":")[3]
        for application in applications:
            if application in j_msg["_source"]["layers"]:
                src_addr = "{:<15}:{:>5}".format(j_msg["_source"]["layers"][ip_layer][ip_layer + ".src"],
                                                 j_msg["_source"]["layers"][transport_layer][
                                                     transport_layer + ".srcport"])
                dst_addr = "{:<15}:{:>5}".format(j_msg["_source"]["layers"][ip_layer][ip_layer + ".dst"],
                                                 j_msg["_source"]["layers"][transport_layer][
                                                     transport_layer + ".dstport"])
                if application == "sip":
                    try:
                        msg_raw = assemble_message_from_json(j_msg, appl=application)
                        msg_obj = buildMessage(msg_raw)
                        msg = check_in_trace(*tests, check_trace=[msg_obj], input_format="list")
                        time_epoch = j_msg["_source"]["layers"]["frame"]["frame.time_epoch"]
                        if not msg:
                            expand = False
                        else:
                            expand = True
                        result[application].setdefault("{}:{}".format(ip_layer, transport_layer), []) \
                            .append((time_epoch, src_addr, dst_addr, msg_obj, expand))
                    except:
                        print("Unable to parse:")
                        pprint(j_msg)
                # result[application]["stream_count"] = len(result[application][ip_layer+":"+transport_layer])
                result[application]["count"] += 1

    return result


def group_continuous_network_traces(file_list, cont_expr=r"(.*)_\d+_\d{8,20}\.(.*)"):
    """Example filename
    <prefix>_00001_20200422141132.<suffix>

    These files are automatically created when capturing wireshark traces with -b

    :file_list: A tuple of filenames or a filename
    :return: A dictionary that groups together files with common prefix and suffix.
            If just one filename is given, return {filename: filename}
    """
    if isinstance(file_list, str):
        return {file_list: file_list}
    file_groups = {}
    for filename in file_list:
        matched = re.match(cont_expr, filename)
        if matched:
            file_group = matched.group(1) + "NNNNN_DDDDDDTTTTTT." + matched.group(2)
            file_groups.setdefault(file_group, []).append(filename)
        else:
            file_groups[filename] = filename
    for group in file_groups:
        if isinstance(file_groups[group], list):
            file_groups[group].sort()
    return file_groups



"""\
Purpose: Simulate a CSTA User
Initial Version: Costas Skarakis 8/7/2020 (Aug 7)
"""
from common.tc_logging import warning, exception
from csta.CstaMessage import is_response, is_event, is_request


class CstaUser:
    def __init__(self, number, csta_application):
        self.number = number
        self.busy = False
        self.csta_application = csta_application
        self.monitorCrossRefID = None
        self.callID = None
        # Transactions are a dictionary of eventid to request eg {353: "MakeCall"}
        self.inc_transactions = {}
        self.out_transactions = {}
        self.deviceID = number
        self.parameters = {"monitorCrossRefID": self.monitorCrossRefID,
                           "CSTA_CREATE_MONITOR_CROSS_REF_ID": self.monitorCrossRefID,
                           "CSTA_USE_MONITOR_CROSS_REF_ID": self.monitorCrossRefID,
                           "callID": self.callID,
                           "deviceID": self.deviceID}

    def get_transaction_id(self, message):
        """ Determine what the correct eventID is for an outgoing message based on the current state and the message.
         EventID and transactionID can be called InvokeID sometimes, like... always"""
        if is_response(message):
            # make sure that this response belongs to an active transaction, for instance if MakeCallResponse is
            # received, make sure there is a MakeCall in our active transactions
            target_request = message.split("Response")[0]
            assert target_request in self.inc_transactions.values(), "{}: Refusing to send {} without having received "\
                                                                 "{} first".format(self.number,
                                                                                   message,
                                                                                   target_request)
            for eventid in self.inc_transactions:
                if self.inc_transactions[eventid] == target_request:
                    # to review if we need to handle parallel transactions of the same type
                    return eventid
        elif is_request(message):
            if not self.out_transactions:
                return self.csta_application.min_event_id
            else:
                return max(int(eventid) for eventid in self.out_transactions) + 1
        else:
            assert is_event(message), "Message {} is not a response, request or event".format(message)
            return 9999

    def update_outgoing_transactions(self, message):
        """ message is of type CstaMessage"""
        if message.is_response():
            if message.eventid in self.inc_transactions:
                self.inc_transactions.pop(message.eventid)
                self.csta_application.min_event_id = max(self.csta_application.min_event_id, message.eventid + 1)
        elif message.is_request():
            if message.eventid in self.out_transactions:
                warning("Sending request {0} with invokeid {1}, "
                        "although request {2} with invokeid {0} has not been answered".format(message.event,
                                                                                              message.eventid,
                                                                                              self.out_transactions[
                                                                                                  message.eventid]))
            self.out_transactions[message.eventid] = message.event
            self.csta_application.min_event_id = max(self.csta_application.min_event_id, message.eventid + 1)
        else:
            assert message.is_event(), "Message {} is not a response, request or event".format(message.event)

    def update_incoming_transactions(self, message):
        """ message is of type CstaMessage"""
        if message.is_response():
            if message.eventid in self.out_transactions:
                self.out_transactions.pop(message.eventid)
                self.csta_application.min_event_id = max(self.csta_application.min_event_id, message.eventid + 1)
            else:
                exception("{}: Received csta response {} with invoke id {} "
                          "but there is no matching active transaction:\n{}\n{}".format(self.number,
                                                                                     message.event,
                                                                                     message.eventid,
                                                                                     self.out_transactions,
                                                                                     message))
        elif message.is_request():
            if message.eventid in self.inc_transactions:
                warning("{0}: Received request {1} with invokeid {2}, while"
                        " request {3} with invokeid {2} is already active".format(self.number,
                                                                                  message.event,
                                                                                  message.eventid,
                                                                                  self.inc_transactions[message.eventid]))
            self.inc_transactions[message.eventid] = message.event
            self.csta_application.min_event_id = max(self.csta_application.min_event_id, message.eventid + 1)
        else:
            assert message.is_event(), "Message {} is not a response, request or event".format(message.event)

    def update_connection_id(self, message):
        if message["deviceID"] and message["callID"]:
            self.callID = message["callID"]
            self.set_parameter("callID", self.callID)

    def monitor_start(self):
        return self.csta_application.monitor_start(self.number)

    def set_parameter(self, key, value):
        self.parameters[key] = value

    def prepare_message(self, message):
        """
        Convert a SIP message string to SipMessage object

        :param message: The message string
        :return: A CstaMessage object
        """
        return self.csta_application.prepare_message(self.number, message)

    def send(self, message, to_user=None):
        return self.csta_application.send(message, from_user=self.number, to_user=to_user)

    def wait_for_message(self, message, ignore_messages=(), timeout=5.0):
        return self.csta_application.wait_for_csta_message(self.number,
                                                           message,
                                                           ignore_messages=ignore_messages,
                                                           timeout=timeout)



"""\
Purpose: Simulate a CSTA Application
Initial Version: Costas Skarakis 7/20/2020  
"""
import os
import traceback
from _socket import timeout as sock_timeout
from threading import Lock
from time import sleep

from common.client import TCPClient
from common.tc_logging import debug, warning, exception
from csta.CstaEndpoint import get_xml
from csta.CstaUser import CstaUser
from csta.CstaParser import parseBytes, buildMessageFromFile, buildMessage


class CstaApplication:
    def __init__(self, server=False):
        self.ip = None
        self.port = None
        self.link = None
        self.min_event_id = 0
        self.users = {}
        self.parameters = {}
        self.server = server
        self.message_buffer = []
        self.waitForCstaMessage = self.wait_for_csta_message  # compatibility alias
        self.lock = Lock()

    def connect(self, local_address, destination_address, protocol="tcp"):
        """ Connect to CSTA Server """
        local_ip, local_port = local_address
        self.ip = local_ip
        # params = {"source_ip": local_ip, "source_port": local_port, "transport": protocol}
        dest_ip, dest_port = destination_address
        # Only TCP implemented
        # 0 means bind to any available local port
        csta_link = TCPClient(local_ip, local_port)
        self.link = csta_link
        csta_link.connect(dest_ip, dest_port)
        self.port = csta_link.port

        inBytes = csta_link.waitForData()
        req = parseBytes(inBytes)

        resp = buildMessageFromFile(get_xml("SystemStatusResponse.xml"), {}, eventid=req.eventid)
        csta_link.send(resp.contents())

        reg = buildMessageFromFile(get_xml("SystemRegister.xml"), {}, eventid=0)
        csta_link.send(reg.contents())

        inBytes = csta_link.waitForData()
        reg_resp = parseBytes(inBytes)
        assert reg_resp.event == "SystemRegisterResponse", 'Invalid Response to System Register Request'
        return csta_link

    def new_user(self, directory_number):
        """ Add a new user to the Application
        :param directory_number: The users number as a string
        :return: The CstaUser object created for this user
        """
        directory_number = str(directory_number)
        if directory_number in self.users:
            warning(directory_number + " already exists")
            return self.users[directory_number]
        else:
            user = CstaUser(directory_number, self)
            self.users[directory_number] = user
            return user

    def get_user(self, directory_number):
        return self.users[str(directory_number)]

    def monitor_start(self, directory_number):
        """ Send MonitorStart and add directory number to monitored users"""
        directory_number = str(directory_number)
        user = self.get_user(directory_number)
        user.parameters["user"] = directory_number
        user.send("MonitorStart")
        inmessage = user.wait_for_message("MonitorStartResponse")
        user.parameters["monitorCrossRefID"] = inmessage["monitorCrossRefID"]

    # def monitor_stop(self, directory_number):
    #     """ Send MonitorStop and delete directory number from monitored users"""
    #     csta_link = self.link
    #     m = buildMessageFromFile(get_xml("MonitorStop.xml"), {"user": directory_number}, eventid=1)
    #     csta_link.send(m.contents())
    #     in_bytes = csta_link.waitForData()
    #     inmessage = parseBytes(in_bytes)
    #     assert inmessage.event == "MonitorStopResponse", "Sent:{}  Received:{}".format(m.event, str(inmessage))
    #     self.parameters.pop(directory_number)

    def get_monitored_users(self):
        return list(user for user in self.users if self.get_user(user).parameters["monitorCrossRefID"] is not None)

    def prepare_message(self, from_user, message):
        """
        Convert a SIP message string to SipMessage object

        :param message: The message string
        :param from_user: The originating CSTA subscriber object
        :return: A CstaMessage object
        """
        user = self.users[from_user]
        return buildMessage(message, user.parameters)

    def send(self, message, from_user=None, to_user=None):
        """ Send a CSTA message

        :param from_user: calling number
        :param to_user: called number
        :param message: This could be a predefined csta message or just the type of the message.
                       There are already several builtin xml files that are named after the corresponding messages
                       In order to access them just provide the type of the CSTA message.
                       There should be a matching filename in csta/xml

                       Example: C.send_csta("MakeCall")
        """
        if isinstance(to_user, CstaUser):
            to_user = to_user.number
        user = self.users[from_user]
        user.set_parameter("callingDevice", from_user)
        user.set_parameter("calledDirectoryNumber", to_user)
        if message.strip().startswith("<?xml"):
            m = buildMessage(message, user.parameters, eventid=0)
        else:
            m = buildMessageFromFile(get_xml(message), user.parameters, eventid=0)
        if from_user and m.event != "MonitorStart":
            assert from_user in self.get_monitored_users(), "User {} must be monitored before sending messages".format(
                from_user)
        with self.lock:
            # must do this serially to avoid sending multiple requests with the same invoke id
            eventid = user.get_transaction_id(m.event)
            m.set_eventid(eventid)
            if m["deviceID"] and m["callID"]:
                m["deviceID"] = user.deviceID
                m["callID"] = user.callID
            user.update_outgoing_transactions(m)
        # old_link = self.link
        # try:
        self.link.send(m.contents())
        # except IOError:
        #     while old_link is self.link:
        #         sleep(0.1)
        #     debug("Retransmitting {} after disconnection and reconnection".format(m.event))
        #     self.link.send(m.contents())

        return m

    def wait_for_csta_message(self, for_user, message, ignore_messages=(), timeout=5.0):
        """
        Wait for CSTA message
        :param for_user: The message intended recipient
        :param message: The message to wait for
        :param ignore_messages: Messages to ignore
        :param timeout: Defined timeout in seconds.
        :return: the CstaMessage received
        """
        inmessage = None
        other_users_xrefid = [usr.parameters["monitorCrossRefID"]
                              if "monitorCrossRefID" in usr.parameters else None
                              for usr in self.users.values() if usr != for_user]
        other_users_transactions = [usr.out_transactions for usr in self.users.values() if usr != for_user]
        user_xrefid = None
        this_user = None
        if for_user is not None:
            this_user = self.users[for_user]
            user_xrefid = this_user.parameters.get("monitorCrossRefID", None)

        while not inmessage:
            self.lock.acquire()
            if self.message_buffer:
                # first get a message from the buffer
                inmessage = self.get_buffered_message(user_xrefid)

            sleep(0.1)

            if self.server:
                continue

            if not inmessage:
                # no (more) buffered messages. try the network
                try:
                    # print("Waiting for", message)
                    # inbytes = self.link.waitForCstaData(timeout=0.2)
                    inbytes = self.link.waitForCstaData(timeout=timeout)
                    inmessage = parseBytes(inbytes)
                # except sock_timeout:
                #     inmessage = None
                #     timeout -= 0.2
                #     if timeout <= 0:
                #         raise
                #     else:
                #         continue
                except UnicodeDecodeError:
                    debug("Ignoring malformed data")
                    inmessage = None
                    continue
                except sock_timeout:
                    exception(str(for_user) + " No " + str(message) + " " + str(self.message_buffer))
                    raise

            inmessage_type = inmessage.event
            if inmessage_type in ignore_messages:
                inmessage = None
                continue

            if message and (
                    # received message is not of expected type or
                    (isinstance(message, str) and message != inmessage_type) or

                    # received message type is not in list of expected types or
                    (type(message) in (list, tuple) and not any([m == inmessage_type for m in message])) or

                    # received message is a csta response that has an unknown/incorrect invokeID (eventid)
                    (this_user is not None and
                     inmessage.is_response() and
                     inmessage.eventid not in this_user.out_transactions) or

                    # received message type is for another user (different xref_id and not MonitorStartResponse)
                    # # ie we are expecting message for a specific user
                    (this_user is not None and
                     # # the message has a xrefid tag and our user has an xrefid parameter (although it may be empty)
                     inmessage["monitorCrossRefID"] and user_xrefid is not None and
                     # # the message xrefid is different than our user's xrefid
                     inmessage["monitorCrossRefID"] != user_xrefid and
                     # # the message is not a MonitorStartResponse. This is the normal case where we send MonitorStart
                     # # at which time the user will not have xrefid assigned yet and we expect a response that we will
                     # # use to assign xrefid to the user
                     inmessage_type != "MonitorStartResponse")
            ):

                # buffer received message if it is intended for another user or if it is an event that came sooner
                # than expected or if it is a response to a request from another user
                if (
                        (this_user is not None and
                         inmessage["monitorCrossRefID"] and user_xrefid and
                         inmessage["monitorCrossRefID"] in other_users_xrefid) or

                        (inmessage.is_event() and
                         inmessage["monitorCrossRefID"] and user_xrefid and
                         inmessage["monitorCrossRefID"] == user_xrefid) or

                        (this_user is not None and
                         inmessage.is_response() and
                         {inmessage.eventid: inmessage_type.replace("Response", "")} in other_users_transactions)
                ):

                    self.message_buffer.append(inmessage)
                    # debug(
                    #     "BUFFERED MESSAGE '{}' with xrefid '{}' for '{}' because I am '{}' waiting for '{}' in '{}'".format(
                    #         inmessage_type,
                    #         inmessage["monitorCrossRefID"],
                    #         inmessage["deviceID"], for_user, message,
                    #         this_user.parameters["monitorCrossRefID"]))
                    inmessage = None
                else:
                    if this_user is None:
                        raise AssertionError('Got "{}" with callID {} and xrefid {} '
                                             'while expecting "{}"'.format(inmessage_type,
                                                                           inmessage["callID"],
                                                                           inmessage["monitorCrossRefID"],
                                                                           message))
                    else:
                        raise AssertionError('Got "{}" with callID {} and xrefid {} while expecting "{}" with '
                                             'callID {} and xrefid {}. '
                                             'Known transactions are:\n"{}"\n{} '.format(inmessage_type,
                                                                                         inmessage["callID"],
                                                                                         inmessage["monitorCrossRefID"],
                                                                                         message,
                                                                                         this_user.parameters.get(
                                                                                             "callID",
                                                                                             None),
                                                                                         this_user.parameters.get(
                                                                                             "monitorCrossRefID",
                                                                                             None),
                                                                                         this_user.out_transactions,
                                                                                         inmessage))
            self.lock.release()

        if this_user is not None:
            # Evaluate the invoke id
            this_user.update_incoming_transactions(inmessage)
            this_user.update_connection_id(inmessage)
        return inmessage

    def get_buffered_message(self, xref_id):
        """
        Return the first buffered message found for the given monitorCrossReferenceID

        :param xref_id: A monitored user's cross reference ID
        :return: the buffered SipMessage
        """
        msg = None
        for i in range(len(self.message_buffer)):
            message = self.message_buffer.pop(0)
            try:
                message_xref_id = message["monitorCrossRefID"]
            except AttributeError:
                message_xref_id = None
            if message_xref_id is None or xref_id is None or xref_id == message_xref_id:
                msg = message
                break
            else:
                self.message_buffer.append(message)
        return msg

    def update_call_parameters(self, directory_number, inresponse):
        """ Update our parameters based on the given incoming CSTA message """
        try:
            if not inresponse.event.endswith("Event") and not inresponse.event.endswith("Response"):
                self.parameters[directory_number]["last_request_eventid"] = inresponse.eventid
            for key in "deviceID", "monitorCrossRefID", "callID":
                if inresponse[key]:
                    self.parameters[directory_number][key] = inresponse[key]
        except AttributeError:
            self.parameters[directory_number].setdefault("deviceID", None)

    def close(self):
        self.link.socket.close()
        self.link = None



"""\
Purpose: Utility functions and classes
Initial Version: Costas Skarakis 11/11/2018
"""
import random
import string
from concurrent.futures import ThreadPoolExecutor
from itertools import cycle
from threading import Timer, Thread
from time import time, sleep
import re
import io
import xml.etree.ElementTree as ET
import logging.handlers
from common.tc_logging import logger
import traceback


def nowHex():
    """ Current time in sec represented in hex - 4 digits """
    return '{:x}'.format(int(time()))


def randStr(digits):
    """ Returns a random string with this many digits"""
    return ''.join((random.choice(string.ascii_lowercase + string.digits) for n in range(digits)))


def randomCallID():
    return nowHex() + randStr(12)


def randomTag():
    """ TODO Implement RFC 3261 Section 19.3 Tags """
    return nowHex() + randStr(4)


def getLocalIP():
    # TODO
    return "10.4.253.10"


def randomBranch():
    return nowHex() + randStr(24)


def epid(*args):
    """ User args to get a unique epid """
    # if we always seed random generation with the same string, the next random number will be the same
    random.seed(''.join(args))
    return hex(random.getrandbits(32))[2:]


def loop(sequence):
    while True:
        for i in sequence:
            yield i


def pool(sequence, condition=bool):
    """
    Cyclically yields the next member of a sequence unless the specified condition is False

    :param sequence: input sequence, eg a list
    :param condition: a function to be run on the next member eg:
            lambda x: x.registered
    :return: yields the next eligible member
    """
    p = cycle(sequence)
    while True:
        c = next(p)
        if condition(c):
            yield c
        else:
            continue


def next_available_sip(sip_pool):
    """Find the next available sip endpoint from a pool of endpoints"""
    busy = True
    a = None
    while busy:
        a = next(sip_pool)
        busy = a.busy
    if type(a).__name__ == "SipEndpointView":
        a.busy = True
        a.update_text()
        a.update_arrow()
        a.colour("green")
    return a


def make_available_sip(*sip_endpoints):
    for sip_endpoint in sip_endpoints:
        sip_endpoint.busy = False
        if type(sip_endpoint).__name__ == "SipEndpointView":
            sip_endpoint.colour("yellow")


def serverThread(target, *args, **kwargs):
    """ Start a thread """
    ex = ThreadPoolExecutor()
    thread = ex.submit(target, *args, **kwargs)
    return thread


def wait_for_sip_data(sockfile):
    content_length = -1
    data = b""
    data += sockfile.readline()

    while True:
        line = sockfile.readline()
        data += line
        if not line.strip():
            break
        header, value = [x.strip() for x in line.split(b":", 1)]
        if header == b"Content-Length" or header == b"content-length":
            content_length = int(value)

    if content_length > 0:
        data += sockfile.read(content_length)

    if content_length == -1:
        raise NoData("No content length in message")
    return data


class NoData(Exception):
    pass


class XmlBody:
    """
    Very simple representation of an xml document

    Get tags as elements with ["tag_name"] or get_tag("tag_name") eg:
        reason_element = xml_body_1["reason"]
        reason_element = xml_body_1.get_tag("reason")
    Set tag text with assignment to tag element eg:
        xml_body_1["reason"] = "Global failure"
    or assign to the element text attribute:
        xml_body_1["reason"].text = "Global failure"
    Set tag attribute with element.set(attribute, value) eg:
        xml_body_1["internalError"].set("message","foo"))
        xml_body_1.get("internalError").set("message","foo"))

    In case the same tag is appearing multiple times get a list with get_all.
    Then you can modify each appearance by iterating on the elements. eg:
        for element in xml_body_1.get_all("tag_name"):
            if element.get("message") == "foo":
                element.text = "bar"

    The above will change this xml:
    <?xml version="1.0"?>
    <tags xmlns="http://people.example.com">
        <tag_name message="not foo">Text1</tag_name>
        <tag_name message="foo">Text2</tag_name>
    </tags>

    To this xml:
    <?xml version="1.0"?>
    <tags xmlns="http://people.example.com">
        <tag_name message="not foo">Text1</tag_name>
        <tag_name message="foo">bar</tag_name>
    </tags>

    Some of these operations may fail in some cases of XML documents with multiple namespaces (not all cases)
    """

    def __init__(self, xml_content, default_encoding="UTF-8",
                 default_namespace="http://www.ecma-international.org/standards/ecma-323/csta/ed4"):
        self.encoding = default_encoding
        if isinstance(xml_content, bytes):
            xml_encoding = re.search(b"encoding=[\'\"](.*)[\'\"] ?\?>", xml_content)
            if xml_encoding:
                self.encoding = xml_encoding.group(1).decode(encoding=default_encoding)
            self.ns_map = self.parse_map(io.StringIO(xml_content.decode(self.encoding)))
        else:
            xml_encoding = re.search("encoding=[\'\"](.*)[\'\"] ?\?>", xml_content)
            if xml_encoding:
                self.encoding = xml_encoding.group(1)
            self.ns_map = self.parse_map(io.StringIO(xml_content))
        try:
            self.body = xml_content.strip()
        except:
            print(xml_content)
            raise
        try:
            root = ET.fromstring(self.body)
        except:
            print(self.body)
            raise
        self.tree = ET.ElementTree(root)
        self.root = self.tree.getroot()
        ns = re.search("^{(.*)}", root.tag)
        if ns:
            self.namespace = ns.group(1)
        else:
            print("Warning: No namespace defined in message", root.tag)
            self.namespace = default_namespace
        #        self.ns_map = dict(re.findall("xmlns:(?P<name>.+) ?= ?[\'\"]?(?P<ns>[^\'\"]+)[\'\"]?[\s> ]",xml_content))
        ET.register_namespace("", self.namespace)
        self.event = self.root.tag.replace("{" + self.namespace + "}", '')

    def parse_map(self, file):
        """
        http://effbot.org/zone/element-namespaces.htm#parsing-with-prefixes

        :return: root tag of xml
        """
        events = "start", "start-ns", "end-ns"

        root = None
        ns_map = []
        count = 0
        for event, elem in ET.iterparse(file, events):
            if event == "start-ns":
                ns_map.append(("ns%d" % count, elem[1]))
                count += 1
        return dict(ns_map)

    def __getitem__(self, key, position="root"):
        """
         Find a tag using the default namespace
        :param key: The requested tag
        :param position: The element to look below
        :return: The first tag in any namespace
        """
        if position == "root":
            position = self.root
        clause = "{" + self.namespace + "}" + key
        if position.tag == clause:
            return position
        else:
            element = position.find(clause)
            if not element:
                for name in self.ns_map:
                    element = position.find(name + ":" + key, self.ns_map)
                    if element: break
            return element

    def get_tag(self, tag, parent="root"):
        """
        Find a tag using tag name, ignoring namespaces
        :param tag: The name of the tag
        :return:
        """
        if parent == "root":
            parent = self.root
        ch = None
        if parent.tag.endswith("}" + tag):
            return parent
        else:
            for child in parent:
                ch = self.get_tag(tag, parent=child)
                if ch is not None:
                    return ch

    def get_all(self, tag, position="root"):
        """
         Find all tags with this name in all namespaces
        :param key: The requested tag name
        :param position: The element to look below
        :return: A list of all elements with specified tag
        """
        if position == "root":
            position = self.root
        elements = []
        if position.tag.endswith("}" + tag):
            elements.append(position)
        else:
            for child in position:
                ch = self.get_tag(tag, parent=child)
                if ch is not None:
                    elements.append(ch)
        return elements

    def get_child(self, parent, tag):
        """
        Find a tags under given parent tag
        :param parent: A xlm tag element
        :param tag: The requested tag name
        :return: The first element found with the requested tag
        """
        return self.get_tag(tag, parent)

    def get_all_children(self, parent, tag):
        """

        :param parent:
        :param tag:
        :return:
        """
        return self.get_all(tag, position=parent)

    def __setitem__(self, key, value):
        element = self.tree.find("{" + self.namespace + "}" + key)
        element.text = value

    def __repr__(self):
        result = io.BytesIO()
        self.tree.write(result, xml_declaration=self.encoding, encoding=self.encoding)
        result.flush()
        return result.getvalue().decode(encoding=self.encoding)

    def __str__(self):
        return repr(self)


class dict_2(dict):
    """
    Override dictionary getitem, to call items that are callable
    """

    def __getitem__(self, item):
        value = dict.__getitem__(self, item)
        if callable(value):
            return value()
        else:
            return value


class Load(object):
    """
    Start a performance run
    """

    def __init__(self,
                 flow,
                 *flow_args,
                 interval=1.0,
                 quantity=1,
                 duration=0,
                 stopCondition=None):
        self.flow = flow
        self.args = flow_args
        self.interval = interval
        self.quantity = quantity
        self.duration = duration
        self.stopCondition = stopCondition
        self.startTime = time()
        self.active = []
        st_logger = logging.getLogger('Statistics')
        handler = logging.handlers.RotatingFileHandler("Statistics.txt", mode="w", maxBytes=20000000, backupCount=5)
        st_logger.addHandler(handler)
        self.log = st_logger
        self.calls = {"Started": 0, "Finished": 0, "Passed": 0}
        self.loop = LoadThread(target=self._start)
        self.statistics()

    def start(self):
        self.loop.start()

    def _start(self):
        """
        Every :interval seconds, start :quantity flows
        """
        while not (self.stopCondition or not (self.duration < 0 or time() - self.startTime < self.duration)):
            t = time()
            for i in range(self.quantity):
                self.run_next_flow()
                sleep(max((0, self.interval-time()+t)))

        self.stop()

    def stop(self):
        """
        Set the stopCondition to stop so that no new execution will be scheduled.
        The running executions will not be interrupted
        """
        self.stopCondition = True

    def run_next_flow(self):
        c = LoadThread(target=self.flow, args=self.args)
        c.start()
        self.active.append(c)
        self.calls["Started"] += 1

    def monitor(self, max_tracked_errors=30):
        while self.active or not self.stopCondition:
            sleep(0.1)
            for inst in (ins for ins in self.active if not ins.is_alive()):
                try:
                    inst.join()
                except Exception as e:
                    key = str(e)
                    if ": Got" in key:
                        key = key.split('"')[1]
                    key = "Failed: " + key
                    if key in self.calls:
                        self.calls[key] += 1
                    elif len(self.calls) < max_tracked_errors:
                        self.calls[key] = 1
                else:
                    self.calls["Passed"] += 1
                finally:
                    self.active.remove(inst)
                    self.calls["Finished"] += 1
        self.loop.join()

    def statistics(self):
        self.log.info("{}:{}".format(time(), self.calls))
        if not self.stopCondition and (self.duration < 0 or time() - self.startTime < self.duration) or self.active:
            Timer(1, self.statistics).start()
        else:
            print("STOPPING")


class LoadThread(Thread):
    """ I had to make a custom thread class to handle exceptions """

    def run(self):
        self.exc = False
        try:
            super().run()
        except Exception as e:
            logger.debug("Exception in Thread")
            logger.debug(traceback.format_exc())
            self.exc = e
            raise

    def join(self, timeout=None):
        if self.exc:
            raise self.exc
        else:
            super().join(timeout)



"""\
Purpose: Functions for parsing SIP messages and creating SipMessage objects
Initial Version: Costas Skarakis 11/11/2018
"""
import re
from sip.SipMessage import SipMessage

ENCODING = "utf8"
MANDATORY_REQUEST_HEADERS = {"To", "From", "CSeq", "Call-ID", "Max-Forwards"}
SPECIAL_CASE_HEADERS = {"Call-ID", "CSeq", "X-Siemens-OSS"}


def buildMessage(message, parameters={}):
    # remove whitespace from start and end TODO test this change on existing testcases
    message = re.sub("\n +", "\n", message)
    tString = message.strip().format(**parameters)
    # replace new lines
    bString = bytes(tString.replace("\n", "\r\n").replace("\r\r\n", "\r\n") + 2 * "\r\n", encoding=ENCODING)
    sipMessage = parseBytes(bString)
    return sipMessage


def parseBytes(bString, sep="\r\n", encoding=ENCODING):
    # header and body are separated by an empty line
    header, body = bString.decode(encoding).split(sep + sep, maxsplit=1)
    header_lines = header.split(sep)
    request_or_response = header_lines[0]
    headers = {}
    count = 0
    for line in header_lines[1:]:
        k, v = line.split(":", maxsplit=1)
        key = str(k).strip()
        if key in headers:
            count += 1
            key = key + "#{}".format(count)
        headers[key] = v.strip()

    # print (headers.keys())
    titled_headers = dict((h.title(), headers[h]) for h in headers)
    case_switched_headers = {}
    for h in titled_headers:
        case_switched_headers[h] = titled_headers[h]
        for special_h in SPECIAL_CASE_HEADERS:
            if h.capitalize() == special_h.capitalize():
                case_switched_headers.pop(h)
                case_switched_headers[special_h] = titled_headers[h]
                break
    message = SipMessage(case_switched_headers, body)
    # add more elements depending if it is a request or a responses
    resP = re.match(r"^SIP/\d\.?\d? (\d+ .*)$", request_or_response, re.I)
    if resP:
        message.type = "Response"
        message.status = resP.group(1)
        message.status_line = request_or_response
    else:
        reqP = re.match(r"^(\w+) \S+ SIP/\d\.?\d?$", request_or_response, re.I)
        if not reqP:
            raise Exception("Not a valid request or response.. or parse logic error", request_or_response)
            # return None

        # make sure all mandatory headers are present
        capital_mandatory_headers = set(h.capitalize() for h in MANDATORY_REQUEST_HEADERS)
        capital_message_headers = set(h.capitalize() for h in headers.keys())
        missing_headers = capital_mandatory_headers.difference(capital_message_headers)
        # missing_headers = MANDATORY_REQUEST_HEADERS.difference(set(headers.keys()))
        if missing_headers:
            raise Exception("Mandatory headers missing", missing_headers)
            # return None
        message.type = "Request"
        message.request_line = request_or_response
        message.method = reqP.group(1)
    # TODO: Implement body parsing
    return message


if __name__ == "__main__":
    m = b'SIP/2.0 403 Forbidden\r\nWarning: 399 10.2.0.22 "Originating Endpoint is not configured or registered on system. Check provisioning of 3021005533, , 10.2.31.5, 10.2.0.24."\r\nCall-ID: 5bc4d2b1lKza5n\r\nCSeq: 1 OPTIONS\r\nTo: <sip:10.2.0.24:5060>\r\nFrom: <sip:3021005533@10.2.31.5:50080>;tag=snl_5bc4d2b14Y\r\nContent-Length: 0\r\nVia: SIP/2.0/TCP 10.2.31.5:50080;branch=5bc4d2b1cswcPR1cq4nQ\r\n\r\n'
    n = b'SIP/2.0 400 Bad Request\r\nWarning: 399 10.2.0.22 "Request mandatory header is missing or incorrect. Mandatory Header CSEQ-Method mismatch."\r\nVia: SIP/2.0/TCP 10.2.31.5:5080;branch=5bc619b78AKFDlh5mRGL\r\nFrom: "3021005533" <sip:3021005533@10.2.0.22:5060>;tag=snl_5bc619b7OD;epid=SCD0n\r\nCSeq: 1 OPTIONS\r\nCall-ID: 5bc619b7TTYmPW\r\nTo: <sip:10.2.0.22:5060>;tag=snl_PT47YjDdJE\r\nContent-Length: 0\r\n\r\n'
    s = parseBytes(n)
    i = '''
   REGISTER sip:10.0.0.1:5060;transport=UDP SIP/2.0
   CSeq: 0 REGISTER
   Call-ID: 1253634160-1253634161-7867113566_SubA-1253634162
   From: 7867113566 <sip:7867113566@10.0.0.1>;epid=7867113566_SubA;tag=359368062
   To: 7867113566 <sip:7867113566@10.0.0.1>
   User-Agent: 
   Expires: 300
   Contact: <sip: 7867113566@10.0.0.2:3566;lr;transport=UDP>
   Via: SIP/2.0/UDP 10.0.0.2:3566;branch=z9hG4bK359368062.17428.1
   Via: SIP/2.0/UDP 10.0.0.3:3566;branch=z9hG4bK359368062.17428.1
   Via: SIP/2.0/UDP 10.0.0.2:3566;branch=z9hG4bK_STANDARD_reboot_with_Authen_Proxy_3-109_1253634158-1
   Route: <sip:7867113566@10.0.0.2:5060;lr;transport=UDP>
   Route: <sip:7867113566@10.0.0.3:5060;lr;transport=UDP>
   Path: <sip:10.0.0.2:5060;transport=UDP;lr>
   Max-Forwards: 70
   Content-Length: 0
'''
    j = buildMessage(i, {})
    s.make_response_to(j)
    x = '''SIP/2.0 202 Accepted
Via: SIP/2.0/TCP esrp.rnspn1.rnsp.california.ng911:5060;branch=z9hG4bKSEC-4d2a050a-4f2a050a-1-H666LTmfe6
record-route: <sip:10.5.42.58:50202;transport=tcp;gwIP=esrp.lane1.dc1.california.ng911~5060~tcp~tcp-5060-tls-5061;oss=bcf-10.09.00.00-1;ftag=snl_0qe665y4ZJ;lr>
contact: <sip:esrpn1@10.5.42.58:50202;transport=tls>
To: "esrpn1" <sip:esrpn1@esrp.lane1.dc1.california.ng911:5060>;tag=SEC11-20ea8c0-60ea8c0-1-w99HwH86Rdgx
From: <sip:rnspn1@esrp.rnspn1.rnsp.california.ng911:5060>;tag=snl_0qe665y4ZJ
X-Siemens-OSS: OpenScape SBC V10 R9.00.00-1/BCF/THIG
Date: Thu, 14 Jan 2021 08:57:17 GMT
call-id: SEC11-4d2a050a-4f2a050a-1-72tEETFIm13o
cseq: 1235 SUBSCRIBE
expires: 2811
content-length: 0'''
    y = '''NOTIFY sip:rnspn1@esrp.rnspn1.rnsp.california.ng911:5060;transport=tcp SIP/2.0
Via: SIP/2.0/TCP [fd00:10:2:8::4]:5061;branch=z9hG4bKdc13.38795be9c926ccdf9ce6c0ba3a91cee9.0;i=16
max-forwards: 69
contact: <sip:esrpn1@esrp.lane1.dc1.california.ng911:5060;transport=tcp>
to: "rnspn1" <sip:rnspn1@esrp.rnspn1.rnsp.california.ng911:5061;transport=tls>;tag=snl_0qe665y4ZJ
from: "esrpn1" <sip:esrpn1@esrp.lane1.dc1.california.ng911:5061;transport=tls>;tag=SEC11-20ea8c0-60ea8c0-1-w99HwH86Rdgx
Date: Thu, 14 Jan 2021 08:57:17 GMT
call-id: SEC11-4d2a050a-4f2a050a-1-72tEETFIm13o
cseq: 1 NOTIFY
content-type: application/vnd.nena.ElementState+xml
subscription-state: active;expires=2811
event: nena-ElementState
content-length: 220
X-Siemens-OSS: OpenScape SBC V10 R9.00.00-1/BCF/THIG

<?xml version="1.0" encoding="UTF-8" ?>
<?xml-stylesheet type="text/xsl" href="elementState.xsl"?>
<ElementState>
    <State>Normal</State>
    <Reason>The element is operating normally.</Reason>
</ElementState>'''
    # print(s.contents())
    ack = '''    ACK sip:302102433001@10.2.31.5:23001;transport=TCP SIP/2.0
    To: <sip:432001@10.2.31.5>;tag=605b7f97rxqy
    From: "Line31001" <sip:431001@10.3.28.112>;tag=snl_Pn97QmG639
    Call-ID: SEC11-6a1c030a-6b1c030a-1-38N4keA0mC7X
    CSeq: 1235 ACK
    Contact: <sip:431001@10.3.28.112:5060;transport=tcp;maddr=10.3.28.112>
    Via: SIP/2.0/TCP 10.3.28.112:5060;branch=z9hG4bKSEC-6a1c030a-6b1c030a-1-DIXL92Tf3A
    Allow: REGISTER, INVITE, ACK, BYE, CANCEL, NOTIFY, REFER, INFO
    Date: Wed, 24 Mar 2021 18:06:16 GMT
    Max-Forwards: 69
    Content-Length: 0'''
    bye = 'BYE sip:SUB_A_ext5@dest_ip:5060;transport=transport;maddr=dest_ip SIP/2.0\n' \
          'Max-Forwards: 70\n' \
          'From: "Line32001" <sip:SUB_B@local_ip:23001>;tag=SubC0-128_33b5341b194b49304833e5e8354c10e9-0\n' \
          'To: <sip:SUB_A_ext5@dest_ip>;tag=snl_6AL2CobU0a\n' \
          'Call-ID: SEC11-a7d980a-a7d980a-1-REgtx7XF0W45\n' \
          'Supported: timer\n' \
          'User-Agent: optiPoint 410_420/V7 V7 R0.11.0 M5T SIP Stack/4.0.24.24\n' \
          'CSeq: 1 BYE\n' \
          'Via: SIP/2.0/tcp local_ip:23001;branch=z9hG4bK1765806562.2480.1\n' \
          'Content-Length: 0\n' \
          '\n'
    ackm = buildMessage(ack, {})
    print(ackm.contents())
    byem = buildMessage(bye, {})
    print(byem.contents())
    byem.make_response_to(ackm)
    print(byem.contents())





"""\
Purpose: Network connection facilities - mock servers
Initial Version: Costas Skarakis 16/2/2020
"""
import selectors
import socket
import threading
import time
import traceback
import types
import ssl
from copy import copy

import common.client as my_clients
from common import util
from common.tc_logging import debug, warning
from sip.SipParser import parseBytes as parseSip, buildMessage
from csta.CstaApplication import CstaApplication
from csta.CstaParser import parseBytes
from csta.CstaUser import CstaUser
from sip.SipEndpoint import SipEndpoint

# With help from https://github.com/realpython/materials/blob/master/python-sockets-tutorial/multiconn-server.py
from sip.messages import message


class SipServer:
    """
    A simple server to send and receive SIP Messages
    """

    def __init__(self, ip, port, protocol="tcp", certificate=None):
        self.ip = ip
        self.port = port
        self.protocol = protocol
        self.connections = []
        self.continue_serving = False
        self.sel = selectors.DefaultSelector()
        if protocol in ("tcp", "TCP"):
            self.socket = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        elif protocol in ("udp", "UDP"):
            self.socket = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
        elif protocol in ("tls", "TLS"):
            tcp_socket = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
            # for MTLS this might be needed
            self.context = ssl.SSLContext(ssl.PROTOCOL_TLSv1_1)
            self.context.load_verify_locations(certificate)
            self.socket = self.context.wrap_socket(tcp_socket, server_hostname='localhost')

            # context.load_cert_chain('/path/to/certchain.pem', '/path/to/private.key')
        self.server_thread = None
        self.sip_endpoint = SipEndpoint("PythonSipServer")
        self.handlers = {}
        self.handlers_args = {}
        self.threads = []
        self.events = {}
        self.buffers = {}
        self.registered_addresses = {}
        self.active_calls = []
        self.links = []
        self.register_links = {}
        self.wait_for_message = self.sip_endpoint.wait_for_message
        self.wait_for_messages = self.sip_endpoint.wait_for_messages
        self.set_dialog = self.sip_endpoint.set_dialog
        self.save_message = self.sip_endpoint.save_message
        self.set_transaction = self.sip_endpoint.set_transaction
        self.on("NOTIFY", self.notify_ok)
        self.on("OPTIONS", self.options_ok)
        self.lock = threading.Lock()

    def get_dialog_link(self, dialog):
        for known_dialog, link in self.links:
            if known_dialog and (
                    known_dialog == dialog or self.sip_endpoint.get_complete_dialog(known_dialog) == dialog):
                return link
        debug("No link found for dialog {}.\nAvailable dialogs and corresponding links:\n{}".format(dialog, self.links))

    def get_address_link(self, address):
        for known_dialog, link in self.links:
            if "{}:{}".format(link.rip, link.rport) == address:
                return link
        debug(
            "No link found for address {}.\nAvailable dialogs and corresponding links:\n{}".format(address, self.links))

    def remove_address_link(self, sock):
        rm_links = []
        for link in self.links:
            if link[1].socket == sock:
                rm_links.append(link)
        for link in rm_links:
            self.links.remove(link)

    def remove_dialog(self, dialog):
        rm_links = []
        for link in self.links:
            if link[0] == dialog:
                rm_links.append(link)
        for link in rm_links:
            self.links.remove(link)

    def send(self, dialog, *args, **kwargs):
        with self.lock:
            self.sip_endpoint.use_link(self.get_dialog_link(dialog))
            return self.sip_endpoint.send(dialog=dialog, *args, **kwargs)

    def send_new(self, address, *args, **kwargs):
        with self.lock:
            link = my_clients.TCPClient(ip=self.ip, port=0)
            rip, rport = address.split(":")
            link.connect(rip, int(rport))
            self.sip_endpoint.use_link(link)
            return self.sip_endpoint.send_new(*args, **kwargs)

    def reply_to(self, inmessage, *args, **kwargs):
        with self.lock:
            self.sip_endpoint.use_link(self.get_dialog_link(inmessage.get_dialog()))
            return self.sip_endpoint.send_in_ctx_of(inmessage, *args, **kwargs)

    def accept_wrapper(self, sock):
        conn, addr = sock.accept()  # Should be ready to read
        print("accepted connection from", addr)
        conn.setblocking(False)
        data = types.SimpleNamespace(addr=addr, inb=b"", outb=b"")
        events = selectors.EVENT_READ  # | selectors.EVENT_WRITE
        self.connections.append(conn)
        self.sel.register(conn, events, data=data)
        self.make_client(conn, addr)

    def make_client(self, sock, addr):
        local_ip, local_port = sock.getsockname()
        client = None
        if self.protocol in ("tcp", "TCP"):
            client = my_clients.TCPClient(local_ip, local_port)
        elif self.protocol in ("udp", "UDP"):
            client = my_clients.UDPClient(local_ip, local_port)
        elif self.protocol in ("tls", "TLS"):
            client = my_clients.TLSClient(local_ip, local_port, None)
        client.rip = addr[0]
        client.rport = addr[1]
        self.sip_endpoint.use_link(client)
        self.sip_endpoint.link.socket = sock
        #        self.socket.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
        self.sip_endpoint.link.sockfile = sock.makefile(mode='rb')
        self.links.append((None, client))
        # self.links.append(("{}:{}".format(*addr), client))
        return client

    def serve_forever(self):
        self.socket.bind((self.ip, self.port))
        if not self.protocol.upper() == "UDP":
            self.socket.listen()
        else:
            self.make_client(self.socket, (self.ip, self.port))
        self.port = self.socket.getsockname()[1]
        print("listening on", (self.ip, self.port))
        self.socket.setblocking(False)
        self.sel.register(self.socket, selectors.EVENT_READ, data=None)
        self.continue_serving = True
        while self.continue_serving:
            try:
                events = self.sel.select(timeout=60)
                for key, mask in events:
                    if key.data is None and self.protocol.upper() != "UDP":
                        self.accept_wrapper(key.fileobj)
                    else:
                        self.service_connection(key, mask)
            except socket.timeout:
                print("timeout")
                traceback.print_exc()
                continue
            except (IOError, socket.error):
                print("Disconnected. Waiting to reconnect")
                # self.sel.unregister(self.csta_endpoint.link.socket)
                continue
            except KeyboardInterrupt:
                self.shutdown()
            except:
                debug(traceback.format_exc())
                self.remove_address_link(key.fileobj)
                self.sel.unregister(key.fileobj)

        self.sel.unregister(self.socket)
        # self.socket.shutdown(socket.SHUT_RDWR)
        self.socket.close()
        self.sel.close()
        self.shutdown()

    def wait_shutdown(self):
        time.sleep(3)
        for thread in self.threads:
            thread.join()
        self.shutdown()

    def shutdown(self):
        print("Shutting down server. It may take up to 5 seconds")
        self.continue_serving = False
        self.server_thread.join()

    def serve_in_background(self):
        self.server_thread = threading.Thread(target=self.serve_forever)
        self.server_thread.daemon = False
        self.server_thread.start()

    def on(self, incoming_message_type, action, args=()):
        """
        Define a response action to a trigger received message

        :param incoming_message_type: The trigger message
        :param action: A function to call when this message is received.
                    The first argument of this function
        :param args: Optional positional arguments to pass to action starting from 3rd position
        """
        self.handlers[incoming_message_type] = action
        self.handlers_args[incoming_message_type] = args
        self.events[incoming_message_type] = threading.Event()
        self.buffers[incoming_message_type] = []
        t = threading.Thread(target=self.on_loop,
                             args=(incoming_message_type, action, args),
                             daemon=True)
        t.start()
        self.threads.append(t)

    def on_loop(self, incoming_message_type, action, args=()):
        while True:
            self.events[incoming_message_type].wait()
            self.events[incoming_message_type].clear()
            while True:
                try:
                    message = self.buffers[incoming_message_type].pop(0)
                    self.update(message)
                    action(self, message, *args)
                except IndexError:
                    break
                except:
                    traceback.print_exc()

    def update(self, inmessage):
        # last_sent_message = self.sip_endpoint.get_last_message_in(dialog)
        # if last_sent_message:
        #     transaction = last_sent_message.get_transaction()
        self.save_message(inmessage)
        if inmessage.type == "Request":
            self.set_transaction(inmessage.get_transaction())
        else:
            self.sip_endpoint.update_to_tag(inmessage.get_dialog())
        self.set_dialog(inmessage.get_dialog())

    def set_parameter(self, key, value, *args):
        """
        Set sip parameters
        :param key: key
        :param value: value assigned to key
        """
        self.sip_endpoint.parameters[key] = value

    def service_connection(self, key, mask):
        sock = key.fileobj
        data = key.data
        address = "{}:{}".format(*data.addr)
        if mask & selectors.EVENT_READ:
            try:
                link = self.get_address_link(address)
                inbytes = self.sip_endpoint.link.waitForSipData(timeout=5.0, client=link)
                if inbytes is None:
                    self.sip_endpoint.link.socket.close()
                    self.sel.unregister(self.sip_endpoint.link.socket)
                    return
                inmessage = parseSip(inbytes)
                in_dialog = inmessage.get_dialog()
                if not in_dialog["to_tag"] and {"Call-ID": in_dialog["Call-ID"],
                                                "from_tag": in_dialog["from_tag"]} in self.sip_endpoint.dialogs:
                    self.sip_endpoint.dialogs.append(in_dialog)
                self.links.append((in_dialog, link))
                if inmessage.get_status_or_method() in self.handlers:
                    self.events[inmessage.get_status_or_method()].set()
                    self.buffers[inmessage.get_status_or_method()].append(inmessage)
                else:
                    self.sip_endpoint.message_buffer.append(inmessage)
            except UnicodeDecodeError:
                debug("Ignoring malformed data")
        # if mask & selectors.EVENT_WRITE:
        #     # print("ready to write")
        #     if data.outb:self.events['MonitorStart']
        #         print("echoing", repr(data.outb), "to", data.addr)
        #         sent = sock.send(data.outb)  # Should be ready to write
        #         data.outb = data.outb[sent:]

    def is_registered(self, user):
        return user.split("@")[0] in self.registered_addresses

    def get_active_call(self, dialog):
        for call in self.active_calls:
            if dialog == call[0]:
                return call[1]
            elif dialog == call[1]:
                return call[0]
        print(dialog, "not in:", self.active_calls)
        return None

    @staticmethod
    def options_ok(self, inmessage):
        self.reply_to(inmessage, message["200_OK_1"])
        self.remove_dialog(inmessage.get_dialog())

    @staticmethod
    def notify_ok(self, inmessage):
        notify_response = buildMessage(message["200_OK_Notify"])
        notify_response.make_response_to(inmessage)
        notify_response["Contact"] = inmessage["Contact"]
        self.reply_to(inmessage, notify_response.contents())


class CstaServer(SipServer):
    def __init__(self, ip, port, protocol="tcp"):
        super().__init__(ip, port, protocol)
        # remove the parent class definition before I redefine it below... not sure why I had to do that
        del self.send
        self.name = "PythonCstaServer"
        self.refid = 0
        self.csta_endpoint = CstaApplication(server=True)
        self.user = self.csta_endpoint.new_user(self.name)
        self.user.parameters["monitorCrossRefID"] = 9999
        self.csta_endpoint.parameters = {self.name: {"eventid": 1}}
        self.wait_for_csta_message = self.csta_endpoint.wait_for_csta_message
        self.wait_for = self.csta_endpoint.wait_for_csta_message
        self.on("MonitorStart", self.monitor_user)
        self.on("SystemStatus", self.system_status)
        self.on("SystemRegister", self.system_register)
        self.on("SnapshotDevice", self.snapshot_device)

    def cleanup(self):
        del self.csta_endpoint
        self.csta_endpoint = CstaApplication(server=True)
        self.user = self.csta_endpoint.new_user(self.name)
        self.user.parameters["monitorCrossRefID"] = 9999
        self.csta_endpoint.parameters = {self.name: {"eventid": 1}}
        self.wait_for_csta_message = self.csta_endpoint.wait_for_csta_message
        self.wait_for = self.csta_endpoint.wait_for_csta_message

    def send(self, message, to_user, from_user=None):
        """ We are using the client methods for the server, which make the from_user and to_user
        terms to have opposite values. We use this method to reverse them and make the code make more sense.

        :param to_user: Will be turned to from_user in CstaApplication.send
        :param from_user: Will be turned to to_user
        :param message: The Message type
        :return: The CstaMessage sent
        """
        if isinstance(to_user, CstaUser):
            return to_user.send(message)
        else:
            return self.csta_endpoint.send(from_user=to_user, message=message, to_user=from_user)

    def update(self, message):
        user = self.csta_endpoint.get_user(self.name)
        user.update_incoming_transactions(message)

    def set_parameter(self, key, value, user=None):
        if not user:
            user = self.user
        self.csta_endpoint.parameters[user][key] = value

    def accept_wrapper(self, sock):
        self.cleanup()
        conn, addr = sock.accept()  # Should be ready to read
        print("accepted connection from", addr)
        conn.setblocking(False)
        data = types.SimpleNamespace(addr=addr, inb=b"", outb=b"")
        events = selectors.EVENT_READ  # | selectors.EVENT_WRITE
        self.connections.append(conn)
        self.make_client(conn, addr)
        self.sel.register(conn, events, data=data)
        self.user.parameters["systemStatus"] = "enabled"
        self.user.parameters["sysStatRegisterID"] = self.name
        self.csta_endpoint.send(from_user=self.name, to_user=None, message="SystemStatus")
        # self.csta_endpoint.wait_for_csta_message(for_user=self.name, message="SystemStatusResponse")
        self.user.parameters["eventid"] = 0

    def make_client(self, sock, addr):
        local_ip, local_port = sock.getsockname()
        client = None
        if self.protocol in ("tcp", "TCP"):
            client = my_clients.TCPClient(local_ip, local_port)
        elif self.protocol in ("udp", "UDP"):
            client = my_clients.UDPClient(local_ip, local_port)
        elif self.protocol in ("tls", "TLS"):
            client = my_clients.TLSClient(local_ip, local_port, None)
        client.rip = addr[0]
        client.rport = addr[1]
        self.csta_endpoint.link = client
        self.csta_endpoint.link.socket = sock
        #        self.socket.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
        self.csta_endpoint.link.sockfile = sock.makefile(mode='rb')
        return client

    def service_connection(self, key, mask):
        # sock = key.fileobj
        # data = key.data
        if mask & selectors.EVENT_READ:
            try:
                inbytes = self.csta_endpoint.link.waitForCstaData(timeout=5.0)
                if inbytes is None:
                    self.csta_endpoint.link.socket.close()
                    self.sel.unregister(self.csta_endpoint.link.socket)
                    return
                inmessage = parseBytes(inbytes)
                if inmessage.event in self.handlers:
                    self.events[inmessage.event].set()
                    self.buffers[inmessage.event].append(inmessage)
                elif inmessage.event == "SystemStatusResponse":
                    # TODO: Ignoring incoming SystemStatusResponse for now
                    return
                else:
                    self.csta_endpoint.message_buffer.append(inmessage)
            except UnicodeDecodeError:
                debug("Ignoring malformed data")
        # if mask & selectors.EVENT_WRITE:
        #     # print("ready to write")
        #     if data.outb:self.events['MonitorStart']
        #         print("echoing", repr(data.outb), "to", data.addr)
        #         sent = sock.send(data.outb)  # Should be ready to write
        #         data.outb = data.outb[sent:]

    def get_user(self, directory_number):
        return self.csta_endpoint.get_user(directory_number)

    @staticmethod
    def monitor_user(csta_server, monitor_message):
        user = monitor_message["deviceObject"]
        if not user:
            debug("Invalid monitor user '{}'. Empty deviceID tag.".format(user))
            return
        csta_server.csta_endpoint.new_user(user).parameters = {"monitorCrossRefID": csta_server.refid,
                                                               "deviceID": user,
                                                               "CSTA_CREATE_MONITOR_CROSS_REF_ID": csta_server.refid,
                                                               "CSTA_USE_MONITOR_CROSS_REF_ID": csta_server.refid}
        csta_server.refid += 1
        csta_server.csta_endpoint.get_user(user).update_incoming_transactions(monitor_message)
        csta_server.send("MonitorStartResponse", to_user=user)
        csta_server.csta_endpoint.get_user(user).update_outgoing_transactions(monitor_message)

    @staticmethod
    def system_register(csta_server, message):
        csta_server.csta_endpoint.parameters[csta_server.name]["sysStatRegisterID"] = csta_server.name
        csta_server.csta_endpoint.send(from_user=csta_server.name, to_user=None, message="SystemRegisterResponse")
        # user.update_outgoing_transactions(message)

    @staticmethod
    def system_status(csta_server, message):
        csta_server.csta_endpoint.parameters[csta_server.name]["systemStatus"] = "normal"
        csta_server.csta_endpoint.send(from_user=csta_server.name, to_user=None, message="SystemStatusResponse")

    #        user.update_outgoing_transactions(message)

    @staticmethod
    def snapshot_device(csta_server, snapshot_device_message):
        user = snapshot_device_message["snapshotObject"]
        csta_server.csta_endpoint.get_user(user).update_incoming_transactions(snapshot_device_message)
        if not user:
            debug("Invalid snapshot device user '{}'. Empty snapshotObject tag.".format(user))
            return
        csta_server.send("SnapshotDeviceResponse", to_user=user)
        csta_server.csta_endpoint.get_user(user).update_outgoing_transactions(snapshot_device_message)



message = {"Options_1": '''\
OPTIONS sip:{dest_ip}:{dest_port};transport={transport} SIP/2.0
Call-ID: {callId}
CSeq: 1 OPTIONS
To: <sip:{dest_ip}:{dest_port}>
From: <sip:{user}@{source_ip}:{source_port}>;tag=snl_{fromTag}
User-Agent: OpenScape Voice V9R0
Content-Length: 0
Max-Forwards: 70
Via: SIP/2.0/{transport} {source_ip}:{source_port};branch={viaBranch}
''',
		"200_OK_Notify":'''
SIP/2.0 200 OK
Contact: Will be overwritten by incoming NOTIFY
Call-ID: Will be overwritten by incoming NOTIFY
CSeq: Will be overwritten by incoming NOTIFY
From: Will be overwritten by incoming NOTIFY
To: Will be overwritten by incoming NOTIFY
Via: Will be overwritten by incoming NOTIFY
Content-Type: application/sdp
Content-Length: 0
''',
        "Subscribe_secondary":'''\
SUBSCRIBE sip:{user}@{dest_ip}:{dest_port};transport={transport} SIP/2.0
Call-ID: {callId}
CSeq: 1 SUBSCRIBE
To: <sip:{user}@{dest_ip}:{dest_port}>
From: "{primary}" <sip:{user}@{dest_ip}:{dest_port}>;tag=snl_{fromTag}
User-Agent: optiPoint 420 Advance/V7 V7 R0.16.1/10.2.31.5
Content-Length: 0
Max-Forwards: 70
Via: SIP/2.0/{transport} {source_ip}:{primary_port};branch={viaBranch}
Accept: application/keyset-info+xml
Allow: NOTIFY
Expires: {expires}
Event: keyset-info
Contact: "{primary}" <sip:{user}@{source_ip}:{primary_port};transport={transport}>
''',
         "Register_secondary":'''\
REGISTER sip:{dest_ip}:{dest_port};transport={transport} SIP/2.0
Call-ID: {callId}
CSeq: 1 REGISTER
To: <sip:{user}@{dest_ip}:{dest_port}>
From: "{user}" <sip:{user}@{dest_ip}:{dest_port}>;tag=snl_{fromTag}
User-Agent: optiPoint 420 Advance/V7 V7 R0.16.1/10.2.31.5
Content-Length: 0
Max-Forwards: 70
Via: SIP/2.0/{transport} {source_ip}:{source_port};branch={viaBranch}
Accept: application/dls-contact-me
Supported: X-Siemens-Proxy-State
Contact: "{user}" <sip:{user}@{source_ip}:{primary_port};transport={transport}>;expires={expires}
''',
         "Register_primary":'''\
REGISTER sip:{dest_ip}:{dest_port};transport={transport} SIP/2.0
Call-ID: {callId}
CSeq: 1 REGISTER
To: <sip:{user}@{dest_ip}:{dest_port}>
From: "{user}" <sip:{user}@{dest_ip}:{dest_port}>;tag=snl_{fromTag};epid={epid}
User-Agent: optiPoint 420 Advance/V7 V7 R0.16.1/10.2.31.5
Content-Length: 0
Max-Forwards: 70
Via: SIP/2.0/{transport} {source_ip}:{source_port};branch={viaBranch}
Accept: application/dls-contact-me
Supported: X-Siemens-Proxy-State
Contact: "{user}" <sip:{user}@{source_ip}:{source_port};transport={transport}>;expires={expires}
''',

           "Register_1": '''\
REGISTER sip:{dest_ip}:{dest_port};transport={transport} SIP/2.0
Call-ID: {callId}
CSeq: 1 REGISTER
To: <sip:{user}@{dest_ip}:{dest_port}>
From: "{user}" <sip:{user}@{dest_ip}:{dest_port}>;tag=snl_{fromTag}
User-Agent: Python tools
Content-Length: 0
Max-Forwards: 70
Via: SIP/2.0/{transport} {source_ip}:{source_port};branch={viaBranch}
Accept: application/dls-contact-me
Supported: X-Siemens-Proxy-State
Contact: "{user}" <sip:{user}@{source_ip}:{source_port};transport={transport}>;expires={expires}
''',
           "Register_2": '''\
REGISTER sip:{dest_ip}:{dest_port};transport={transport} SIP/2.0
Call-ID: {callId}
CSeq: 1 REGISTER
To: <sip:{user}@{dest_ip}:{dest_port}>
From: "{user}" <sip:{user}@{dest_ip}:{dest_port}>;tag=snl_{fromTag};epid={epid}
User-Agent: Python tools
Content-Length: 0
Max-Forwards: 70
Via: SIP/2.0/{transport} {source_ip}:{source_port};branch={viaBranch}
Accept: application/dls-contact-me
Supported: X-Siemens-Proxy-State
Contact: "{user}" <sip:{user}@{source_ip}:{source_port};transport={transport}>;expires={expires}
''',
           "Invite_SDP_1": '''\
INVITE sip:{userB}@{dest_ip}:{dest_port};transport={transport} SIP/2.0
Via: SIP/2.0/{transport} {source_ip}:{source_port};branch={viaBranch}
From: {userA} <sip:{userA}@{dest_ip}:{dest_port}>;tag={fromTag}
To: <sip:{userB}@{dest_ip}:{dest_port};transport={transport}>
Contact: <sip:{userA}@{source_ip}:{source_port};transport={transport}>
Content-Type: application/sdp
Call-ID: {callId}
CSeq: 1 INVITE
Max-Forwards: 70
Content-Length: Will be overridden during creation

v=0
o=Anomymous {userA} 1234567890 IN IP4 {source_ip}
s=SIGMA is the best
c=IN IP4 {source_ip}
t=0 0
m=audio 6006 RTP/AVP 8 0 3
a=rtpmap:8 PCMA/8000
a=rtpmap:0 PCMU/8000
a=rtpmap:3 GSM/8000
m=video 6008 RTP/AVP 40
a=rtpmap:40 H263-1998/90000
''',
           "Trying_1": '''\
SIP/2.0 100 Trying
Call-ID: Will be overwritten by incoming INVITE
CSeq: Will be overwritten by incoming INVITE
From: Will be overwritten by incoming INVITE
To: Will be overwritten by incoming INVITE
Via: Will be overwritten by incoming INVITE
Content-Length: 0
''',
           "Ringing_1": '''\
SIP/2.0 180 Ringing
Call-ID: Will be overwritten by incoming INVITE
CSeq: Will be overwritten by incoming INVITE
From: Will be overwritten by incoming INVITE
To: Will be overwritten by incoming INVITE and a tag will be added
Via: Will be overwritten by incoming INVITE
Contact: <sip:{user}@{source_ip}:{source_port};transport={transport}>
Content-Length: 0
''',
           "200_OK_SDP_1": '''\
SIP/2.0 200 OK
Contact: <sip:{user}@{source_ip}:{source_port};transport={transport}>
Call-ID: Will be overwritten by incoming INVITE
CSeq: Will be overwritten by incoming INVITE
From: Will be overwritten by incoming INVITE
To: Will be overwritten by incoming INVITE and a tag will be added
Via: Will be overwritten by incoming INVITE
Content-Type: application/sdp
Content-Length: Will be overridden during creation

v=0
o=Anomymous {user} 1234567890 IN IP4 {source_ip}
s=SIGMA is the best
c=IN IP4 {source_ip}
t=0 0
m=audio 6006 RTP/AVP 8 0 3
a=rtpmap:8 PCMA/8000
a=rtpmap:0 PCMU/8000
a=rtpmap:3 GSM/8000
m=video 6008 RTP/AVP 40
a=rtpmap:40 H263-1998/90000
''',
           "Ack_1": '''\
ACK sip:{userB}@{dest_ip}:{dest_port};transport={transport};maddr={dest_ip} SIP/2.0
CSeq: 1 ACK
Via: SIP/2.0/{transport} {source_ip}:{source_port};branch={viaBranch}
To: Same as initial INVITE
From: Same as initial INVITE
Call-ID: Same as initial INVITE
Max-Forwards: 70
Content-Length: 0
''',
           "Bye_1": '''\
BYE sip:{userB}@{dest_ip}:{dest_port};transport={transport};maddr={dest_ip} SIP/2.0
Call-ID: Same as initial INVITE
CSeq: 2 BYE
Via: SIP/2.0/{transport} {source_ip}:{source_port};branch={viaBranch}
To: Same as initial INVITE
From: Same as initial INVITE
Max-Forwards: 70
Content-Length: 0
''',
           "200_OK_1": '''\
SIP/2.0 200 OK
Content-Length: 0
Call-ID: Will be overwritten by incoming BYE
CSeq: Will be overwritten by incoming BYE
From: Will be overwritten by incoming BYE
To: Will be overwritten by incoming BYE
Via: Will be overwritten by incoming BYE
''',
           "Notify_terminated_1": '''\
NOTIFY sip:{user}@{dest_ip}:{dest_port};transport={transport} SIP/2.0
Via: SIP/2.0/{transport} {source_ip};branch={viaBranch}
Max-Forwards: 70
From: {user} <sip:{user}@{source_ip}:{source_port};transport={transport}>
To: <sip:{user}@{dest_ip}:{dest_port};transport={transport}>
Call-ID: {callId}
CSeq: 465101717 NOTIFY
Contact: <sip:{user}@{source_ip}:{source_port};transport={transport}>
Event: keyset-info
Subscription-State: active;expires={expires}
User-Agent: OpenStage_60_V3 R1.41.0      SIP  130205
Content-Type: application/keyset-info+xml
Content-Length: 316

<?xml version="1.0"?>
<keyset-info xmlns="urn:ietf:params:xml:ns:keyset-info"
	version="0"
	entity="sip:{user}@{dest_ip}:{dest_port}">
	<ki-data> <ki-state>"unconfirmed"</ki-state> <ki-event>"unknown"</ki-event>  </ki-data>
	<di:dialog id="no-dialog"> 
		<di:state>terminated</di:state>
	</di:dialog>
</keyset-info>
''',
           "487_Request_Terminated": '''\
SIP/2.0 487 Request Terminated
Via: SIP/2.0/{transport} {dest_ip}:5060;branch={viaBranch}
From: <sip:{user}@{dest_ip}>;tag={fromTag}
To: {user} <sip:{user}@{source_ip}:{source_port}>
Call-ID: {callId}
CSeq: 1235 INVITE
Server: optiPoint 420 Standard/V7 V7 R2.1.0
Content-Length: 0
'''
           }



"""\
Purpose: CSTA message object
Initial Version: Costas Skarakis 11/11/2018
"""
import os
import re
import xml.etree.ElementTree as ET
import io

xmlpath = os.path.join(os.path.realpath(__file__).replace(os.path.basename(__file__), ''), "CstaPool")


def is_response(message):
    if message.endswith("Response"):
        return True
    else:
        return False


def is_request(message):
    if is_event(message) or is_response(message):
        return False
    else:
        return True


def is_event(message):
    return message.endswith("Event")


class CstaMessage(object):
    """
    Representation of a CSTA message
    """

    def __init__(self, header, xml_tree, str_body=None, encoding="UTF-8",
                 ns=(("", "http://www.ecma-international.org/standards/ecma-323/csta/ed4"),)):
        # s_ip,s_port,d_ip,d_port
        self.eventid = int(header[-4:])
        self.size = header[:4]
        self.header = header
        self.encoding = encoding
        self.str_body = str_body
        self.body = xml_tree
        self.namespace = ""
        self.root = self.body.getroot()
        for namespace in ns:
            if not re.match(r"ns\d+", namespace[0]):
                # namespace is pair of (name, url)
                ET.register_namespace(*namespace)
            if namespace[0] == "":
                # default namespace
                self.namespace = namespace[1]
#        self.event = self.root.tag.replace("{" + self.namespace + "}", '')
        self.event = re.match(r"({.*})?(.*)", self.root.tag).group(2)

    def is_response(self):
        return is_response(self.event)

    def is_request(self):
        return is_request(self.event)

    def is_event(self):
        return is_event(self.event)

    def find_element(self, key):
        element = None
        for tag in self.body.iter():
            element = tag.find(key)
            if element is None:
                element = tag.find("{" + self.namespace + "}" + key)
            if element is None:
                element_search = re.search(key, tag.tag)
                if element_search:
                    element = tag
            if element is not None:
                break
        return element

    def __getitem__(self, key):
        element = self.find_element(key)
        if element is None:
            return None
        else:
            return element.text

    def __setitem__(self, key, value):
        element = self.find_element(key)
        element.text = value

    def __repr__(self):
        result = io.BytesIO()
        self.body.write(result,
                        xml_declaration=self.encoding,
                        encoding=self.encoding)
        result.flush()
        return result.getvalue().decode(encoding=self.encoding)

    def __str__(self):
        return repr(self)

    def message(self):
        return str(self)

    def set_eventid(self, eventid):
        self.header = self.header[:4] + bytes("%04d" % eventid, encoding=self.encoding)
        self.eventid = eventid

    def contents(self):
        """ returns byte string """
        # Recalculate length
        self.size = bytes.fromhex("%08X" % (len(self.message()) + 8))
        return self.size + \
               self.header[-4:] + \
               bytes(self.message(), encoding=self.encoding)


if __name__ == "__main__":
    print(is_response("GetAgentStateResponse"))
    print(is_request("GetAgentStateResponse"))


import argparse
from datetime import datetime
import pprint
from os import chdir, listdir
from os import path
import sys
sys.path.append("..")
sys.path.append(path.join("..", ".."))
from tshark_tools.lib import summarize_trace


def analyze(tracefiles, *criteria, hide_unmatched=False, wireshark_filter=None):
    if isinstance(tracefiles, str):
        tracefiles = (tracefiles,)
    output = "Expanding messages matching:\n"
    for criterion in criteria:
        output += pprint.pformat(criterion)
        output += "\n"
    output += "\n"
    outfile = ""
    calls = []
    for tracefile in tracefiles:
        output += "\nTraceFile: " + tracefile + "\n"
        if tracefile.endswith("json"):
            i_format = "json"
        else:
            i_format = "pcapng"
        print(criteria)
        summary = summarize_trace(tracefile, *criteria, input_format=i_format, tshark_filter=wireshark_filter)

        for transport in summary["sip"]:
            if isinstance(summary["sip"][transport], list):
                for time_epoch, fromaddr, toaddr, message, expand in summary["sip"][transport]:
                    timestamp = datetime.fromtimestamp(float(time_epoch)).strftime("%Y-%m-%dT%H:%M:%S.%f")
                    try:
                        callid = message["Call-ID"]
                    except:
                        print(message.contents())
                        raise
                    if callid not in calls:
                        calls.append(callid)
                    call_number = calls.index(callid) + 1
                    m_lines = message.contents().split("\r\n")
                    o_line = "{} | Call#{:0>4}: {:20} {:-^14}> {:20} | {}\n".format(timestamp, call_number, fromaddr, transport, toaddr, m_lines[0])
                    if expand or not hide_unmatched:
                        output += o_line
                    if expand:
                        for m_line in m_lines[1:]:
                            output += ("{:" + str(len(o_line.rsplit("|", 1)[0])+2) + "}{}\n").format(" ", m_line)
        if not outfile:
            outfile = tracefile.rsplit(".", 1)[0]+".txt"
            mode = "w"
        else:
            mode = "a"
        with open(outfile, mode) as sip_trace:
            sip_trace.write(output)
            output = ""
    return outfile


def GetArgs():
    """
    Supports the command-line arguments listed below.
    """
    parser = argparse.ArgumentParser(description='Summarize network capture file with expanded')
    parser.add_argument('-d', '--directory', required=False, action='store',
                        help='The directory containing the files to analyze')
    parser.add_argument('-f', '--filter', required=False, action='store',
                        help='Wireshark filter to apply to output')
    parser.add_argument('-m', '--message', required=False, action='store',
                        help='Expand SIP messages whose Request URI or Status Code contain this string')

    args = parser.parse_args()
    return args


if __name__ == "__main__":
    # diff(ref_trace="reference.json", check_trace="traceoutput.json", input_format="json")
    ignore = {"Message": ["OPTIONS"]}

    # diff(ref_trace="simclientsip_side.pcapng", check_trace="bcfe.pcap", input_format="pcapng", filters=ignore)

    Test0 = {"xml": {"any": "ns:recording",
                     "label tag": "audio",
                     "nameID aor attr": "sip:912119@grtsep41l3n1sipsm.h8k.sec"}
             }

    TestKat1 = {"Message": "INVITE",
                "Headers": {"Call-Info": "eidd"}
                }
    TestKat2 = {"Message": "REFER",
                "Headers": {"Refer-To": "eidd"},
                }

    Test1 = {"Headers": {"From": "302200880895"}
             }
    Test2 = {"Message": "480"
             }
    # check_in_trace(TestKat1, TestKat2, check_trace="bcf_e_selective.pcap")

    #ignore = {"Message": ["OPTIONS", "SUBSCRIBE"]}
    # diff(ref_trace="bcf_egress_comfort_noise_enabled.pcap", check_trace="bcf_e_selective.pcap", input_format="pcapng", filters=ignore)

    # pprint(summarize_trace("bcf_egress_comfort_noise_enabled.pcap"))

    args = GetArgs()
    chdir(args.directory)
    for file in listdir("."):
        if file.rsplit(".", 1)[-1] in ("pcapng", "pcap", "cap", "json"):
            analyze(file, Test1, Test2, wireshark_filter=args.filter)



"""\
Purpose: Simulate a CSTA device
Initial Version: Costas Skarakis 29/4/2019  
"""
import os
from _socket import timeout

from common import util
from common.client import TCPClient
from csta.CstaMessage import xmlpath
from sip.SipEndpoint import SipEndpoint
from csta.CstaParser import parseBytes as parseBytes_csta, buildMessageFromFile, buildMessage
import sip.SipFlows as SipFlows


def get_xml(name):
    if not name.endswith(".xml"):
        name += ".xml"
    return os.path.join(xmlpath, name)


class CstaEndpoint(SipEndpoint):
    """
    SipEndpoint with CSTA capabilities
    """

    def __init__(self, directory_number):
        super().__init__(directory_number)
        self.csta_links = []
        self.incoming_event_threads = []
        self.eventid = 1
        self.last_sent_csta_message = None
        self.parameters["deviceID"] = None
        self.parameters["callID"] = None
        self.waitForCstaMessage = self.wait_for_csta_message  # compatibility alias

    def sip_connect(self, local_address, destination_address, protocol="tcp"):
        """ Wrap SipEndpoint.connect with a different name """
        return self.connect(local_address, destination_address, protocol)

    def csta_connect(self, local_address, destination_address, protocol="tcp"):
        """ Connect to CSTA Server """
        if not "source_ip" in self.parameters:
            #raise Exception("Must connect SIP first")
            local_ip, local_port = local_address
            self.ip = local_ip
            self.port = local_port
            self.parameters["source_ip"] = local_ip
            self.parameters["source_port"] = local_port
            self.parameters["transport"] = protocol
        else:
            local_ip = self.parameters["source_ip"]
            local_port = 0
        dest_ip, dest_port = destination_address
        # Only TCP implemented
        # 0 means bind to any available local port
        csta_link = TCPClient(local_ip, local_port)
        self.csta_links.append(csta_link)
        csta_link.connect(dest_ip, dest_port)

        inBytes = csta_link.waitForData()
        req = parseBytes_csta(inBytes)

        resp = buildMessageFromFile(get_xml("SystemStatusResponse.xml"), self.parameters, eventid=req.eventid)
        csta_link.send(resp.contents())

        reg = buildMessageFromFile(get_xml("SystemRegister.xml"), self.parameters, eventid=0)
        csta_link.send(reg.contents())

        inBytes = csta_link.waitForData()
        reg_resp = parseBytes_csta(inBytes)
        assert reg_resp.event == "SystemRegisterResponse", 'Tried to start a new dialog with a SIP Response'
        return csta_link

    def monitor_start(self):
        """ Connect a CSTA client link, send MonitorStart """
        csta_link = self.csta_links[0]
        m = buildMessageFromFile(get_xml("MonitorStart.xml"), self.parameters, eventid=1)
        csta_link.send(m.contents())
        in_bytes = csta_link.waitForData()
        inmessage = parseBytes_csta(in_bytes)
        assert inmessage.event == "MonitorStartResponse", "Sent:{}  Received:{}".format(m.event, str(inmessage))

    def start_new_incoming_event_thread(self, csta_link=None, role=None):
        """ Start a new thread to wait for incoming CSTA events

        :role optional argument that can be used to impose certain checks on incoming messages
            Possible values: 1) Any of "originator", "initiator", "A", "Aside", "caller" or
                             2) Any of "destination", "target", "B", "Bside", "callee"

            If any of the group 1) role is selected the incoming messages will be checked to be in the following order:
              "ServiceInitiatedEvent", "OriginatedEvent", "DeliveredEvent", "EstablishedEvent","ConnectionClearedEvent"

            If any of the group 2) role is selected the incoming messages will be checked to be in the following order:
              "DeliveredEvent", "EstablishedEvent", "ConnectionClearedEvent"

            If no role is selected there will be no checks for incoming csta messages

            TODO: Add support for custom message ordering check
        """
        if not csta_link:
            csta_link = self.csta_links[0]
        self.incoming_event_threads.append(util.serverThread(self.wait_for_csta_events,
                                                             csta_link,
                                                             role))

    def wait_for_csta_events(self, csta_link, assert_leg=None):
        """ Waits for CSTA messages until the CSTA link is not valid """
        if not csta_link:
            csta_link = self.csta_links[0]
        if assert_leg in ("originator", "initiator", "A", "Aside", "caller"):
            assert_message_queue = ["ServiceInitiatedEvent", "OriginatedEvent", "DeliveredEvent", "EstablishedEvent",
                                    "ConnectionClearedEvent"]
        elif assert_leg in ("destination", "target", "B", "Bside", "callee"):
            assert_message_queue = ["DeliveredEvent", "EstablishedEvent", "ConnectionClearedEvent"]
        else:
            assert_message_queue = []

        # reset_queue = assert_message_queue[:]

        while self.csta_links:
            # Close and unset csta_link to stop the thread
            try:
                in_bytes = csta_link.waitForCstaData(timeout=1.0)
                inmessage = parseBytes_csta(in_bytes)
                # if assert_leg and not assert_message_queue:
                    # Reset queue in case it needs to be reused
                    # assert_message_queue = reset_queue[:]
                if assert_message_queue:
                    expected_event = assert_message_queue.pop(0)
                    assert inmessage.event == expected_event, \
                        "User {} expected {} but got {}".format(self.parameters["user"], expected_event,
                                                                inmessage.event)
                # print("User:{} received {}".format(user,inmessage))
            except timeout:
                pass

    def send_new_sip(self, target_sip_ep=None, message_string="", expected_response=None, ignore_messages=()):
        """ Wrapper for SipEndpoint.send_new. Starts a new SIP dialog and send a SIP Request """
        return self.send_new(target_sip_ep, message_string, expected_response, ignore_messages)

    def send_sip(self, message_string="", expected_response=None, ignore_messages=(), dialog=None):
        """ Wrapper for SipEndpoint.send. Send a SIP message within an existing SIP dialog """
        return self.send(message_string, expected_response, ignore_messages, dialog)

    def reply_sip(self, message_string, dialog=None):
        """ Wrapper for SipEndpoint.reply Send a response to a previously received message """
        return self.reply(message_string, dialog)

    def send_csta(self, message_xml_file, target=None):
        """ Send a CSTA message

        :message_xml_file: This could be a predefined csta message or just the type of the message.
                           There are already several builtin xml files that are named after the corresponding messages
                           In order to access them just provide the type of the CSTA message.
                           There should be a matching filename in csta/xml

                           Example: C.send_csta("MakeCall")
        """
        self.eventid += 1
        if target and (isinstance(target, SipEndpoint) or isinstance(target, CstaEndpoint)):
            self.parameters["callingDevice"] = self.parameters["userA"] = self.number
            self.parameters["calledDirectoryNumber"] = self.parameters["userB"] = target.number
        if message_xml_file.strip().startswith("<?xml"):
            m = buildMessage(message_xml_file, self.parameters, self.eventid)
        else:
            m = buildMessageFromFile(get_xml(message_xml_file), self.parameters, self.eventid)
        self.last_sent_csta_message = m
        self.csta_links[0].send(m.contents())
        return m

    def wait_for_csta_message(self, message_type, ignore_messages=(), new_dialog=False, timeout=5.0):
        """
        Wait for CSTA message
        :param message_type: The message to wait for
        :param ignore_messages: Messages to ignore
        :param new_dialog: If False, the incoming session ID must be the same as the one of the message we last sent (same dialog)
        :param timeout: Defined timeout in seconds.
        :return: the CstaMessage received
        """
        inmessage = None
        while not inmessage or inmessage.event in ignore_messages:
            in_bytes = self.csta_links[0].waitForCstaData(timeout=timeout)
            inmessage = parseBytes_csta(in_bytes)
        assert inmessage.event == message_type, \
            '{}: Got "{}" while expecting "{}"'.format(self.number, inmessage.event, message_type)
        if inmessage.eventid != 9999 and not new_dialog:
            assert inmessage.eventid == self.eventid, \
                '{}: Wrong event id received: {} \n' \
                'Event id expected: {}\n' \
                'Last message sent: \n{}' \
                '\nMessage received:\n{}\n'.format(self.number,
                                                   inmessage.eventid,
                                                   self.eventid,
                                                   self.last_sent_csta_message,
                                                   inmessage)
            self.update_call_parameters(inmessage)
        return inmessage

    def update_call_parameters(self, inresponse):
        """ Update our deviceID based on the given incoming CSTA response """
        try:
            self.parameters["deviceID"] = inresponse["deviceID"]
            self.parameters["callID"] = inresponse["callID"]
        except AttributeError:
            self.parameters["deviceID"] = None
            self.parameters["callID"] = None

    def unregister(self):
        """ Unregister SIP and stop all incoming CSTA event threads """
        super().unregister()
        for csta_link in self.csta_links:
            csta_link.socket.close()
        self.csta_links = []
        for thread in self.incoming_event_threads:
            thread.result()



"""\
Purpose: Testcase Logging
Initial Version: Costas Skarakis 11/11/2018
"""
import logging.config
import logging.handlers

LOG_CONFG = {
    'version': 1,
    'disable_existing_loggers': True,
    'formatters': {
        'verbose': {
            'format': '%(asctime)s %(levelname)s %(message)s'
        },
        'simple': {
            'format': '%(levelname)s %(message)s'
        }
    },
    'handlers': {
        'console': {
            'level': 'ERROR',
            'class': 'logging.StreamHandler',
            'formatter': 'simple'
        },
        'logfile': {
            'level': 'DEBUG',
            'mode': 'w',
            'class': 'logging.FileHandler',
            'formatter': 'verbose',
            'filename': r'CallTrace.txt'
        },
    },
    'root': {
        'handlers': ['console', 'logfile'],
        'level': 'DEBUG'
    }
}

logging.config.dictConfig(LOG_CONFG)
logger = logging.getLogger(__name__)
debug, info, warning, error, critical, exception = \
    logger.debug, logger.info, logger.warning, logger.error, logger.critical, logger.exception



"""\
Purpose: Network connection facilities
Initial Version: Costas Skarakis 21/6/2019
"""
import sys
from queue import Empty
import socket, select
from traceback import print_exc

from common.tc_logging import debug
import multiprocessing as mp
from sip.SipParser import parseBytes as parseSip, buildMessage as buildSip


class Server(object):
    def __init__(self, server_type, local_addr, remote_addr, in_queue, internal_socket, events_queue):
        self.type = server_type
        self.local_addr = local_addr
        self.remote_addr = remote_addr
        self.queue = in_queue
        self.internal_socket = internal_socket
        self.internal_socket.listen()
        self.events_queue = events_queue
        self.sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        self.sock.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
        self.sock.bind(local_addr)
        self.sock.settimeout(5.0)
        self.sock.connect(remote_addr)
        self.sockfile = self.sock.makefile(mode='rb')
        self.expected_messages = []
        self.lock = mp.Lock()
        try:
            self.serve()
        except KeyboardInterrupt:
            self.queue.put("Stop")

    def consume_sip_data(self):
        try:
            #debug("reading for sip data")
            content_length = -1
            data = b""
            data += self.sockfile.readline()

            while True:
                line = self.sockfile.readline()
                data += line
                if not line.strip():
                    break
                header, value = [x.strip() for x in line.split(b":", 1)]
                if header == b"Content-Length":
                    content_length = int(value)

            if content_length > 0:
                data += self.sockfile.read(content_length)

            if content_length == -1:
                #debug(data.decode())
                #raise Exception TODO
                print("No content length in message")
        except ValueError:
            debug(data.decode())
            debug(line.decode())
            print_exc()
            #raise
        except socket.timeout:
            #debug('Data received before timeout: "{}"'.format(data.decode()))
            print_exc()
            #raise
        self.handle_sip_message(data)

    def consume_csta_data(self, timeout=None):
        header = self.sock.recv(4)
        datalength = int(''.join(["%02X" % x for x in header]), base=16) - 4
        data = b''
        while len(data) < datalength:
            data += self.sock.recv(datalength - len(data))
        # print('message:\n{}\nsize{}\n'.format(data,datalength))
        return header + data

    def handle_sip_message(self, data, encoding="utf8"):
        self.update_sip_event_handling()
        inmessage = parseSip(data)
        message = inmessage.get_status_or_method()
        dialog = inmessage.get_dialog()
        for msg, dlg, resp in self.expected_messages:
            # If the dialog dlg is {} we will respond to all msg messages received
            if msg == message and (dlg == dialog or dlg == {}):
                if resp:
                    response = buildSip(resp, {})
                    response.make_response_to(inmessage)
                    response.set_transaction_from(inmessage)
                    self.lock.acquire()
                    self.sock.sendall(bytes(response.contents(), encoding))
                    self.lock.release()
                # If no response is given for these messages then they will just be ignored
                return
        self.queue.put(data)

    def update_sip_event_handling(self):
        try:
            # incoming queue object should be <Method or Status>, dialog, response message <string template>
            message, dialog, response = self.events_queue.get_nowait()
            self.expected_messages.append((message, dialog, response))
        except Empty:
            return

    def serve(self):
        inputs = [self.sock, self.internal_socket]
        outputs = []
        message_queues = {}
        while inputs:
            #sleep(0.01)
            readable, writable, exceptional = select.select(inputs, outputs, inputs)
            if exceptional:
                raise Exception("Exception in select")
            for r in readable:
                if r is self.sock:
                    if self.type == "sip":
                        self.consume_sip_data()
                    elif self.type == "csta":
                        self.consume_csta_data()

                elif r is self.internal_socket:
                    connection, client_address = self.internal_socket.accept()
                    inputs.append(connection)
                    message_queues[connection] = mp.Queue()
                else:
                    data = r.recv(1024)
                    self.lock.acquire()
                    self.sock.sendall(data)
                    self.lock.release()

            # for s in writable:
            #     try:
            #         next_msg = message_queues[s].get_nowait()
            #     except Empty:
            #         outputs.remove(s)
            #     else:
            #         self.lock.acquire()
            #         s.sendall(next_msg)
            #         self.lock.release()
            #
            # for s in exceptional:
            #     inputs.remove(s)
            #     if s in outputs:
            #         outputs.remove(s)
            #     s.close()


class TCPClient(object):
    def __init__(self, ip, port):
        self.ip = ip
        self.port = port
        self.rip = None
        self.rport = None
        self.rip, self.rport = None, None
        self.lock = mp.Lock()
        self.in_queue = mp.Queue()
        self.out_queue = mp.Queue()
        self.internal_socket = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        self.internal_socket.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
        self.internal_socket.bind(("127.0.0.1", 0))
        self.internal_client = socket.socket()
        self.events_queue = mp.Queue()
        self.server = None
        # self.buffer = []
        self.endpoints_connected = 0

    def connect(self, dest_ip, dest_port):
        self.rip = dest_ip
        self.rport = dest_port
        self.server = self.start_server("sip")

    def start_server(self, server_type):
        """ Start a server process """
        server = mp.Process(target=Server, name=str("{}:{}".format(self.ip, self.port)), args=(server_type,
                                                                                         (self.ip,self.port),
                                                                                         (self.rip, self.rport),
                                                                                         self.in_queue,
                                                                                         self.internal_socket,
                                                                                         self.events_queue))
        server.start()
        self.internal_client.connect(self.internal_socket.getsockname())
        return server

    def register_for_event(self, event):
        """ Event should be three-tuple of <Method or Status>, dialog, response message <string template> """
        self.events_queue.put(event)

    def send(self, data, encoding="utf8"):
        # try:
        #     data = self.out_queue.get_nowait()
        # except Empty:
        #     return
        if type(data) == type(b''):
            debug("Sending from port {}:\n\n".format(self.port) + data.decode("utf8", "backslashreplace").replace(
                "\r\n", "\n"))
            self.internal_client.sendall(data)
        else:
            debug("Sending from port {}:\n\n".format(self.port) + data.replace("\r\n", "\n"))
            self.internal_client.sendall(bytes(data, encoding))

    def waitForSipData(self, timeout=5):
        if not self.server:
            self.start_server("sip")
        try:
            return self.wait_for_data(timeout)
        except Empty:
            raise TimeoutError("Timeout waiting for message on {}:{}".format(self.ip, self.port))

    def waitForCstaData(self, timeout=5):
        if not self.server:
            self.start_server("csta")
        return self.wait_for_data(timeout)

    def wait_for_data(self, timeout=None):
        debug("Waiting on address {}:{}".format(self.ip, self.port))
        try:
            data = self.in_queue.get(timeout=timeout)
            if data == "Stop":
                raise KeyboardInterrupt
        except Empty:
            debug("No message received after timeout")
            raise
        if isinstance(data, tuple):
            dbg = data[0]
        else:
            dbg = data
        debug("Received on address {}:{}:\n\n".format(self.ip, self.port) + dbg.decode("utf8").replace("\r\n", "\n"))
        return data



class UDPClient(TCPClient):
    def __init__(self, ip, port):
        self.ip = ip
        self.port = port
        self.rip, self.rport = None, None
        self.socket = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
        self.socket.bind((self.ip, self.port))
        self.socket.settimeout(5.0)
        self.sockfile = self.socket.makefile(mode='rb')

if __name__ == "__main__":
    from sip.messages import message
    lip = "10.114.72.154"
    A = TCPClient(lip, 41141)
    A.connect(lip, 8000)
    A.send(b'''GET / HTTP/1.1
Host: localhost:8000
User-Agent: Mozilla/5.0 (Windows NT 6.1; Win64; x64; rv:67.0) Gecko/20100101 Firefox/67.0
Accept: text/html,application/xhtml+xml,application/xml;q=0.9,*/*;q=0.8
Accept-Language: en-US,en;q=0.5
Accept-Encoding: gzip, deflate
Connection: keep-alive
Upgrade-Insecure-Requests: 1

''')
    A.register_for_event(("200 OK", {}, message["200_OK_Notify"]))
    A.waitForSipData()



"""\
Purpose: SIP message object
Initial Version: Costas Skarakis 11/11/2018
"""

import re
import hashlib
from common import util


def get_user_from_message(sip_message, header=None):
    pattern = r"<?sip:([^;>]*).*[>;]"
    if not header:
        header_value = sip_message.request_line
    else:
        header_value = sip_message[header]
    M = re.search(pattern, header_value)
    if M:
        return M.group(1)
    else:
        return None


class SipMessage(object):
    """
    Representation of a SIP message
    """

    def __init__(self, header_dict, body):
        self.header = header_dict
        self.headers = self.header
        self.status = None
        self.method = None
        self.type = None
        self.status_line = ""
        self.request_line = ""
        self.body = body
        if body:
            self.header["Content-Length"] = str(len(body.strip()) + 4)
        else:
            self.header["Content-Length"] = "0"
        self.to_tag = ""
        self.from_tag = ""
        self.via_branch = ""
        if "To" in self.header:
            M = re.search(r"tag=([^;]+)", self.header["To"])
            if M:
                self.to_tag = M.group(1)
        if "From" in self.header:
            M = re.search(r"tag=([^;]+)", self.header["From"])
            if M:
                self.from_tag = M.group(1)
        if "Via" in self.header:
            M = re.search(r"branch=([^;]+)", self.header["Via"])
            if M:
                self.via_branch = M.group(1)

    def __eq__(self, message_or_method_or_status):
        """
        Override equality operator to return comparison only based on the method or status string

        :param message_or_method_or_status: The method or status string in the request line of the message
        :return: True or False
        """
        if isinstance(message_or_method_or_status, str):
            other_status_or_method = message_or_method_or_status
        elif isinstance(message_or_method_or_status, SipMessage):
            other_status_or_method = message_or_method_or_status.get_status_or_method()
        else:
            raise Exception("Can only compare SipMessage to str or SipMessage, "
                            "not {}".format(type(message_or_method_or_status)))

        return self.get_status_or_method() == other_status_or_method

    def __getitem__(self, key):
        for k in self.header:
            # handle different letter case
            if k.upper() == key.upper():
                return self.header[k]
        raise KeyError("%s not in message headers" % key)

    def __setitem__(self, key, value):
        self.header[key] = value

    def __repr__(self):
        if self.from_tag:
            if "tag" in self.header["From"]:
                self.header["From"] = re.sub("tag=[^;]+", "tag=%s", self.header["From"]) % self.from_tag
            else:
                self.header["From"] = self.header["From"] + ";tag=" + self.from_tag

        if self.to_tag:
            if "tag" in self.header["To"]:
                self.header["To"] = re.sub("tag=[^;]+", "tag=%s", self.header["To"]) % self.to_tag
            else:
                self.header["To"] = self.header["To"] + ";tag=" + self.to_tag
        else:
            if "tag" in self.header["To"]:
                self.header["To"] = self.header["To"].split(";tag=")[0]

        if self.via_branch:
            if "branch" in self.header["Via"]:
                self.header["Via"] = re.sub("branch=[^;]+", "branch=%s", self.header["Via"]) % self.via_branch
            else:
                self.header["Via"] = self.header["Via"] + ";branch=" + self.via_branch

        if self.type == "Request":
            first_line = self.request_line
        else:
            first_line = self.status_line
        result = first_line + "\r\n"
        result += "\r\n".join(k.split("#")[0] + ": " + v for k, v in self.header.items())
        result += "\r\n"
        result += "\r\n" + self.body
        return result

    def get_transaction(self):
        try:
            cseq, method = self.header["CSeq"].split()
        except:
            cseq, method = self.header["cseq"].split()
        return {"via_branch": self.via_branch,
                "cseq": cseq,
                "method": method
                }

    def get_dialog(self):
        return {"Call-ID": self["Call-ID"],
                "to_tag": self.to_tag,
                "from_tag": self.from_tag}

    def get_dialog_string(self):
        """

        :return: Dialog string in CallID:FromTag:ToTag format
        """
        return "%s:%s:%s" % (self["Call-ID"], self.from_tag, self.to_tag)

    def set_dialog_from(self, other):
        """
        Set current dialog elements from another sip message, or dialog string, or dialog dict
        :SipMessage other: The message/dict/string to get the dialog elements from
        """
        if type(other) == type(self):
            self.from_tag = other.from_tag
            self.to_tag = other.to_tag
            self.header["Call-ID"] = other["Call-ID"]
        elif isinstance(other, dict):
            self.from_tag = other["from_tag"]
            self.to_tag = other["to_tag"]
            self.header["Call-ID"] = other["Call-ID"]
        elif isinstance(other, str):
            self.header["Call-ID"], self.from_tag, self.to_tag = other.split(":")
        else:
            raise Exception("Cannot set SipMessage dialog from %s" % type(other))

    def set_transaction_from(self, other):
        """
        Set current transaction elements from another sip message, or dialog string, or dialog dict
        :SipMessage other: The message/dict/string to get the transaction elements from
        """
        if type(other) == type(self):
            self.via_branch = other.via_branch
            method = self.method if self.method else other.method
            cseq = self.header["CSeq"].split()[0]
            self.header["CSeq"] = " ".join([cseq, method])
        elif isinstance(other, dict):
            self.via_branch = other["via_branch"]
            method = self.method if self.method else other["method"]
            self.header["CSeq"] = " ".join([other["cseq"], method])
        elif isinstance(other, str):
            self.header["CSeq"], self.via_branch = other.split(":")
        else:
            raise Exception("Cannot set SipMessage transaction from %s" % type(other))

    def make_response_to(self, other, dialog={}):
        """ RFC 3261 Section 8.2.6.2 Headers and Tags
        :dict dialog: The current sip dialog we are in, represented by
            a dictionary with keys as in the return type of self.get_dialog()
        """
        self.set_dialog_from(other)
        self.via_branch = other.via_branch
        self["From"] = other["From"]
        self["Call-ID"] = other["Call-ID"]
        if self.type == "Response":
            # Update Via only when we send a response.
            # Otherwise adjust this request for the current dialog
            self["Via"] = other["Via"]
        # else:
        #     self["Via"] = "SIP/2.0/{} {}:{};branch={}".format(self.proto, self.ip, self.port, self.via_branch)
        self["To"] = other["To"]
        if not other.to_tag:
            # This is useful when we make a response to an initial invite.
            # The invite has no to tag. But we made one for Trying
            # and now we must use it for Ringing as well.
            if "to_tag" in dialog and dialog["to_tag"]:
                self.to_tag = dialog["to_tag"]
            else:
                self.to_tag = util.randomTag()
        return self.get_dialog()

    def __str__(self):
        return repr(self)

    def message(self):
        return repr(self)

    def contents(self):
        return self.message()

    def get_status_or_method(self):
        """" Return what kind of message this is. Handy for assertions """
        if self.type == "Request":
            return self.method
        elif self.type == "Response":
            return self.status

    def increase_cseq(self):
        cseq, method = self.header["CSeq"].split()
        self["CSeq"] = " ".join([str(int(cseq) + 1), method])

    def addAuthorization(self, indata, user, pwd):
        """indata is the WWW-Authenticate header we received"""
        self.increase_cseq()
        md5_a1hash = hashlib.md5()
        md5_a2hash = hashlib.md5()
        md5_responsehash = hashlib.md5()

        data = [x.split("=") for x in indata.split(',')]
        D = dict((x[0].strip(), x[1].strip()) for x in data)
        Uri = re.search("sip:(.*) SIP", self.request_line)
        User = user
        RemoteIP = Uri.group(1)
        NcStr = "00000001"  # str(int(time.time()))
        RealmStr = D["Digest realm"][1:-1]
        NonceStr = D["nonce"][1:-1]
        CnonceStr = "970b9994"
        UriStr = "sip:" + RemoteIP
        md5_a1hash.update(bytes(User + ":" + RealmStr + ":" + pwd, encoding="utf8"))
        MD5_A1Str = md5_a1hash.hexdigest()
        md5_a2hash.update(bytes(self.method + ":" + UriStr, encoding="utf8"))
        MD5_A2Str = md5_a2hash.hexdigest()
        md5_responsehash.update(bytes(MD5_A1Str + ":" +
                                      NonceStr + ":" +
                                      NcStr + ":" +
                                      CnonceStr + ":" +
                                      "auth" + ":" +
                                      MD5_A2Str, encoding="utf8"))
        MD5_ResponseStr = md5_responsehash.hexdigest()
        authorization = 'Digest response="{}",'
        authorization += 'nc={},'
        authorization += 'username="{}",'
        authorization += 'realm="{}",'
        authorization += 'nonce="{}",'
        authorization += 'algorithm=MD5,'
        authorization += 'qop=auth,'
        authorization += 'cnonce="{}",'
        authorization += 'uri="{}"'
        authorization = authorization.format(MD5_ResponseStr,
                                             NcStr,
                                             User,
                                             RealmStr,
                                             NonceStr,
                                             CnonceStr,
                                             UriStr)
        self.__setitem__("Authorization", authorization)


if __name__ == "__main__":
    pass



"""\
Purpose: Simulate a SIP phone/line appearance/user
Initial Version: Costas Skarakis
"""
from threading import Timer

from common.tc_logging import exception
from sip.SipParser import parseBytes, buildMessage
import common.util as util
import common.client as client
import sip.SipFlows as flow
from threading import Lock
from sip.SipMessage import SipMessage


def dialog_hash(dialog):
    return "|".join([dialog["Call-ID"], dialog["from_tag"], dialog["to_tag"]])


class SipEndpoint(object):
    """\
    Representation of a SIP Endpoint
    """

    def __init__(self, directory_number):
        """
        Set some initial attribute values
        """
        self.ip = None
        self.port = None
        self.link = None
        self.number = directory_number
        self.parameters = {"user": directory_number,
                           "callId": None,
                           "fromTag": None,
                           "viaBranch": None,
                           "epid": None,
                           "expires": 360,
                           "cseq": None
                           }
        self.last_messages_per_dialog = []
        self.dialogs = []
        self.requests = []
        self.current_dialog = {
            "Call-ID": None,
            "from_tag": None,
            "to_tag": None
            # "epid": None,
        }
        self.tags = {}
        self.current_transaction = {"via_branch": None,
                                    "cseq": "0",
                                    "method": None
                                    }

        self.waitForMessage = self.wait_for_message  # compatibility alias
        self.reply_to = self.send_in_ctx_of  # method alias
        self.secondary_lines = []
        self.message_buffer = []
        self.registered = False
        self.re_register_timer = None
        self.busy = False
        self.lock = Lock()

    def update_parameters(self, params, force=False):
        """
        Update endpoint parameters. This is useful for adding more flexibility in the creation of SIP messages

        :param params: The key-value pairs input
        :param force: Force update of existing variables
        :return: None
        """
        for parameter in params:
            if parameter not in self.parameters or force is True:
                self.parameters[parameter] = params[parameter]

    def connect(self, local_address, destination_address, protocol="tcp", certificate=None, subject_name="localhost"):
        """ Connect to the SIP Server """
        local_ip, local_port = local_address
        dest_ip, dest_port = destination_address

        self.parameters["dest_ip"] = dest_ip
        self.parameters["dest_port"] = dest_port
        self.parameters["transport"] = protocol
        if protocol in ("tcp", "TCP"):
            self.link = client.TCPClient(local_ip, local_port)
        elif protocol in ("udp", "UDP"):
            self.link = client.UDPClient(local_ip, local_port)
        elif protocol in ("tls", "TLS"):
            # for MTLS this might be needed
            # context.load_cert_chain('/path/to/certchain.pem', '/path/to/private.key')
            self.link = client.TLSClient(local_ip, local_port, certificate)
        else:
            raise NotImplementedError("{} client not implemented".format(protocol))
        self.link.connect(dest_ip, dest_port)
        self.set_address((local_ip, self.link.port))

    def update_to_tag(self, in_dialog):
        """
        Update the to_tag on an existing dialog that has no to_tag
        :param in_dialog: the complete dialog
        :return: None
        """
        with self.lock:
            for dialog in self.dialogs:
                if dialog["Call-ID"] == in_dialog["Call-ID"] \
                        and dialog["from_tag"] == in_dialog["from_tag"] \
                        and not dialog["to_tag"]:
                    dhash = dialog_hash(dialog)
                    dialog["to_tag"] = in_dialog["to_tag"]
                    dhash_complete = dialog_hash(dialog)
                    self.tags[dhash_complete] = self.tags[dhash]

    def get_complete_dialog(self, in_dialog):
        """
        Look for a complete existing dialog in case the provided one doesn't have a to tag
        :param in_dialog: the possibly incomplete dialog
        :return: in_dialog if it is complete or a corresponding complete one is not found,
                 otherwise the existing corresponding complete dialog
        """
        if in_dialog["to_tag"]:
            return in_dialog
        with self.lock:
            for dialog in self.dialogs:
                if dialog["Call-ID"] == in_dialog["Call-ID"] \
                        and dialog["from_tag"] == in_dialog["from_tag"] \
                        and dialog["to_tag"]:
                    return dialog
            return in_dialog

    def set_address(self, address):
        """
        We can create and Endpoint just to use it as target for sip messages

        :param address: The address as an (ip, port) tuple
        :return: None
        """
        local_ip, local_port = address
        self.ip = local_ip
        self.port = local_port
        self.parameters["source_ip"] = local_ip
        self.parameters["source_port"] = local_port

    def use_link(self, link):
        """ Convenience function to use an existing network connection"""
        protocol = ["TCP", "UDP"][link.socket.proto]
        local_ip = link.ip
        local_port = link.port
        dest_ip = link.rip
        dest_port = link.rport
        self.link = link
        self.ip = local_ip
        self.port = local_port
        self.parameters["source_ip"] = local_ip
        self.parameters["source_port"] = local_port
        self.parameters["dest_ip"] = dest_ip
        self.parameters["dest_port"] = dest_port
        self.parameters["transport"] = protocol
        # self.link.endpoints_connected += 1

    def set_dialog(self, dialog):
        """ Change current dialog to the one provided """
        if not isinstance(dialog, dict):
            exception("Must provide a dialog in the form of Python dictionary")
        for key in self.current_dialog:
            if key not in dialog:
                exception("Not a valid dialog. Missing key: " + key)
        dialog = self.get_complete_dialog(dialog)
        if dialog not in self.dialogs:
            self.dialogs.append(dialog)
            self.requests.append([])
            # If we don't know this dialog it means we didn't started so it must be an incoming Request
            self.tags[dialog_hash(dialog)] = "to_tag"
        self.current_dialog = dialog
        self.parameters["callId"] = self.current_dialog["Call-ID"]
        self.parameters["fromTag"] = self.current_dialog["from_tag"]
        self.parameters["toTag"] = self.current_dialog["to_tag"]
        return dialog

    def switch_tags(self, dialog=None):
        if not dialog:
            dialog = self.current_dialog
        dhash = dialog_hash(dialog)
        dialog["from_tag"], dialog["to_tag"] = dialog["to_tag"], dialog["from_tag"]
        dhash_switched = dialog_hash(dialog)
        self.set_dialog(dialog)
        if self.tags[dhash] == "to_tag":
            self.tags[dhash_switched] = "from_tag"
        else:
            self.tags[dhash_switched] = "to_tag"

    def set_transaction(self, transaction):
        """ Change current transaction to the one provided """
        if not isinstance(transaction, dict):
            exception("Must provide a transaction in the form of Python dictionary")
        for key in self.current_transaction:
            if key not in transaction:
                exception("Not a valid transaction. Missing key: " + key)
        for key in self.current_transaction:
            self.current_transaction[key] = transaction[key]

    def start_new_dialog(self):
        """ Refresh the SIP dialog specific parameters """
        dialog = {
            "Call-ID": util.randomCallID(),
            "from_tag": util.randomTag(),
            "to_tag": ""
            # "epid": lambda x=6: "SC" + util.randStr(x),
        }
        self.current_dialog = dialog
        self.tags[dialog_hash(dialog)] = "from_tag"
        self.dialogs.append(dialog)
        self.requests.append([])
        return dialog

    def get_last_message_in(self, dialog):
        """ Get the last message sent or received in the provided dialog """
        # If we have received no messages yet return None
        if self.current_dialog == {"Call-ID": None, "from_tag": None, "to_tag": None}:
            return None
        with self.lock:
            # First check for complete dialogs
            for message in self.last_messages_per_dialog:
                if message.get_dialog() == dialog:
                    return message
            # Second check for incomplete dialogs
            for message in self.last_messages_per_dialog:
                d = message.get_dialog()
                if d["Call-ID"] == dialog["Call-ID"] and d["from_tag"] == dialog["from_tag"]:
                    return message
        raise Exception("No message found in dialog {}. Other dialogs active: ".format(dialog, self.dialogs))

    def save_message(self, message):
        """ Search for previously received message in the same dialog.
            If found, replace with given message, otherwise append message to message list """
        with self.lock:
            for i in range(len(self.last_messages_per_dialog)):
                if message.get_dialog() == self.last_messages_per_dialog[i].get_dialog():
                    self.last_messages_per_dialog[i] = message
                    return
            self.last_messages_per_dialog.append(message)

    def prepare_message(self, message):
        """
        Convert a SIP message string to SipMessage object

        :param message: The message string
        :return: A SipMessage object
        """
        return buildMessage(message, self.parameters)

    def start_new_transaction(self, method, dialog=None):
        """ Refresh the via branch and CSeq header """
        if not dialog:
            dialog = self.current_dialog
        try:
            last_message_in_dialog = self.get_last_message_in(dialog)
        except:
            last_message_in_dialog = None
        branch = util.randomBranch()
        if method in ("ACK", "CANCEL"):
            # Not really a new transaction
            # find transaction from last message in dialog
            # which should not be none because what are we sending ACK or CANCEL to?
            transaction = last_message_in_dialog.get_transaction()
            cseq = transaction["cseq"]
        elif last_message_in_dialog:
            transaction = last_message_in_dialog.get_transaction()
            cseq = str(int(transaction["cseq"]) + 1)
        else:
            with self.lock:
                cseq = str(len(self.requests[self.dialogs.index(dialog)]))
                if method != "REGISTER":
                    # Ignore REGISTER otherwise unregister breaks
                    # TODO: check if reRegistrations will work
                    self.requests[self.dialogs.index(dialog)].append(method)
        transaction = {"via_branch": branch, "cseq": cseq, "method": method}
        self.current_transaction = transaction
        return transaction

    def send_new(self, target_sip_ep=None, message_string="", expected_response=None, ignore_messages=[]):
        """ Start a new dialog and send a message """
        self.parameters["userA"] = self.number
        if target_sip_ep:
            if isinstance(target_sip_ep, SipEndpoint):
                self.parameters["userB"] = target_sip_ep.number
                target_sip_ep.parameters["userB"] = self.number
            elif isinstance(target_sip_ep, str):
                self.parameters["userB"] = target_sip_ep
            else:
                raise Exception("target_sip_ep must be str or SipEndpoint, not {}".format(type(target_sip_ep)))
        else:
            # In cases like REGISTER, there is no B-side
            # Clear parameters to avoid unexpected results
            # self.parameters.pop("userA", None)
            # self.parameters.pop("userB", None)
            pass

        if isinstance(message_string, SipMessage):
            m = message_string
        else:
            m = buildMessage(message_string, self.parameters)
        assert m.type == "Request", 'Tried to start a new dialog with a SIP Response'

        new_dialog = self.start_new_dialog()
        new_transaction = self.start_new_transaction(m.method)  # m should always be a request

        m.set_dialog_from(new_dialog)
        m.set_transaction_from(new_transaction)

        self.link.send(m.contents())
        self.save_message(m)

        if expected_response:
            # try:
            self.waitForMessage(message_type=expected_response, dialog=new_dialog, ignore_messages=ignore_messages)
        # except AssertionError:
        #     raise AssertionError('{}: "{}" response to "{}"\n{}'.format(self.number,
        #                                                                 self.get_last_message_in(
        #                                                                     m.get_dialog()).get_status_or_method(),
        #                                                                 m.method,
        #                                                                 m))
        return m

    def send_in_ctx_of(self, reference_message, this_message_string="", expected_response=None, ignore_messages=[]):
        """ Send a message within the same dialog and transaction as 'reference_message' """
        self.set_transaction(reference_message.get_transaction())
        return self.send(message_string=this_message_string,
                         expected_response=expected_response,
                         ignore_messages=ignore_messages,
                         dialog=reference_message.get_dialog())

    def send(self, message_string="", expected_response=None, ignore_messages=[], dialog=None):
        """ Send a message within a dialog """
        m = self.reply(message_string, dialog)
        if expected_response:
            self.waitForMessage(message_type=expected_response, ignore_messages=ignore_messages, dialog=dialog)
        return m

    def reply(self, message_string, dialog=None):
        """ Send a response to a previously received message """
        if "callId" not in self.parameters or not self.parameters['callId']:
            raise Exception("Cannot reply when we are not in a dialog")

        if isinstance(message_string, SipMessage):
            # We can also get a SipEndpoint type as input but it will be sent as is
            m = message_string
        else:
            m = buildMessage(message_string, self.parameters)

        if dialog:
            dialog = self.set_dialog(dialog)
        else:
            dialog = self.current_dialog

        try:
            previous_message = self.get_last_message_in(dialog)
            m.make_response_to(previous_message)
            self.update_to_tag(m.get_dialog())
        except:  # caused by get_last_message_in(dialog) if no message has been exchanged in this dialog yet
            # New dialog, same call-id, eg NOTIFY after Keyset SUBSCRIBE.
            # Must be a SIP Request
            previous_message = None
            assert m.type == "Request", \
                "Attempted to send a {} response in a new dialog".format(m.get_status_or_method())

        if m.type == "Request":
            # This is a new request in the same dialog, so fix the CSeq
            # m.increase_cseq()
            self.start_new_transaction(m.method)
            # If B side sends the new request we must switch from and to tags
            if previous_message is not None and \
                    m.method not in ("ACK", "CANCEL") and \
                    self.tags[dialog_hash(dialog)] == "to_tag":
                self.switch_tags(dialog)
            m.set_dialog_from(dialog)
        else:
            self.free_resources(m)
        self.save_message(m)

        m.set_transaction_from(self.current_transaction)
        self.link.send(m.contents())
        return m

    def get_buffered_message(self, dialog):
        """
        Return the first buffered message found in the given dialog

        :param dialog: The dialog in question
        :return: the buffered SipMessage
        """
        msg = None
        # for message in self.message_buffer:
        for i in range(len(self.message_buffer)):
            message = self.message_buffer[i]
            d = message.get_dialog()
            # If we have received no messages yet return the first message in the buffer
            if self.current_dialog == {"Call-ID": None, "from_tag": None, "to_tag": None} or \
                    (d["Call-ID"] == dialog["Call-ID"] and d["from_tag"] == dialog["from_tag"]):
                msg = message
                self.message_buffer.pop(i)
                break
        return msg

    def wait_for_message(self, message_type, dialog=None, ignore_messages=(), link=None, timeout=5.0):
        """
        Wait for a specific type of SIP message.
        :param message_type: is a string that we will make sure is
                             contained in the received message request or response line
                             Set this to None or "" to accept any incoming message

                             if a list or tuple is given any of the contained message types
                             will be accepted

        :param dialog: Set the dialog to expect the message in. If None will expect a message in current dialog
        :param ignore_messages: a list of message_types to ignore if they come before
                                the expected message_type
        :param timeout: Defined timeout in seconds.
        :return: A SipMessage constructed from the incoming message
        """
        if not link:
            link = self.link
        if not dialog:
            dialog = self.current_dialog
        else:
            dialog = self.get_complete_dialog(dialog)

        inmessage = None
        last_sent_message = self.get_last_message_in(dialog)
        if last_sent_message:
            transaction = last_sent_message.get_transaction()
        len_buffer = len(self.message_buffer)
        count = 0

        while not inmessage:

            if count < len_buffer:
                # first get a message from the buffer
                inmessage = self.get_buffered_message(dialog)
                count += 1

            if not inmessage:
                # no (more) buffered messages. try the network
                inbytes = self.link.waitForSipData(timeout=timeout, client=link)
                inmessage = self.handleDA(last_sent_message, parseBytes(inbytes))

            inmessage_type = inmessage.get_status_or_method()
            inmessage_dialog = inmessage.get_dialog()
            inmessage.cseq_method = inmessage.get_transaction()["method"]

            if inmessage_type in ignore_messages:
                inmessage = None
                continue

            if message_type and \
                    ((isinstance(message_type, str) and message_type not in inmessage_type) or
                     (type(message_type) in (list, tuple) and not any([m in inmessage_type for m in message_type])) or
                     (inmessage.type == "Response" and inmessage.cseq_method != transaction["method"])):
                # we have received an unexpected message. buffer it if there is an active dialog for it
                if inmessage_dialog in self.dialogs or inmessage_type == "INVITE":
                    # message is part of another active dialog or a new call, so buffer it
                    self.message_buffer.append(inmessage)
                    # print("Appended {} with {} to buffer. Will keep waiting for {} in {} ".format(inmessage_type,
                    #                                                                        inmessage_dialog,
                    #                                                                        message_type,
                    #                                                                            dialog))
                    inmessage = None
                else:
                    d = ["sip:{}@".format(line.number) in inmessage["To"] for line in self.secondary_lines]
                    if any(d):
                        # message is meant for another line in this device
                        self.secondary_lines[d.index(True)].message_buffer.append(inmessage)
                        inmessage = None
                    else:
                        raise AssertionError('{}: Got "{}" in {} while expecting "{}" in {}. '
                                             'Other active dialogs:{}.'.format(self.number,
                                                                               inmessage_type,
                                                                               inmessage_dialog,
                                                                               message_type,
                                                                               dialog,
                                                                               self.dialogs))

        self.save_message(inmessage)
        if inmessage.type == "Request":
            self.set_transaction(inmessage.get_transaction())
        else:
            assert inmessage.cseq_method == transaction["method"], \
                "Got {} to {} instead of {} to {}".format(inmessage.get_status_or_method(),
                                                          inmessage.cseq_method,
                                                          message_type,
                                                          transaction["method"])
            self.update_to_tag(inmessage.get_dialog())
            self.free_resources(inmessage)
        self.set_dialog(inmessage.get_dialog())
        return inmessage

    def wait_for_messages(self, *list_of_message_types, in_order=False, ignore_messages=[]):
        """
        Wait for a list of messages. To be used in a loop or with the next() function.
        Each invocation will return the next SipMessage received.
        The loop will exit when all messages have been received

        :param list_of_message_types: The list of messages expected eg ["NOTIFY", "SUBSCRIBE", "403 Forbidden"]
        :param in_order: Impose the given order for incoming messages
        :param ignore_messages:
        :return: SipMessage received after each invocation
        """
        l = len(list_of_message_types)
        not_received_messages = list(list_of_message_types)
        received_messages = []
        for i in range(l):
            try:
                sip_message = self.wait_for_message(None, ignore_messages)
            except TimeoutError:
                raise TimeoutError("Timeout waiting for {}. Messages already received: {}".format(not_received_messages,
                                                                                                  received_messages))
            message_received = sip_message.get_status_or_method()
            assert message_received in not_received_messages, \
                'After receiving {} we expected one of {} but got {}'.format(received_messages,
                                                                             not_received_messages,
                                                                             message_received)
            if in_order:
                message_order = list_of_message_types.index(message_received)
                assert i == message_order, \
                    '{} arrived in order {} instead of {}'.format(message_received, i, message_order)
            not_received_messages.remove(message_received)
            received_messages.append(message_received)
            yield sip_message

    def set_digest_credentials(self, username, password, realm=""):
        """
        Set the digest authentication credentials
        :param username: digest username
        :param password: digest password
        :param realm: digest realm - no effect at the moment since it is overridden by 401 response realm
        """
        self.parameters.update({
            "da_user": username,
            "da_pass": password,
            "realm": realm
        })

    def handleDA(self, request, response):
        """" Add DA to message and send again """
        # Usual case in lab, password same as username
        if "da_pass" not in self.parameters or "da_user" not in self.parameters:
            self.set_digest_credentials(self.number, self.number, "")
        user, pwd = self.parameters["da_user"], self.parameters["da_pass"]
        if response.type == "Response" and response.status == "401 Unauthorized":
            request.addAuthorization(response["WWW-Authenticate"], user, pwd)
            self.link.send(request.contents())
            inBytes = self.link.waitForSipData()
            return parseBytes(inBytes)
        else:
            return response

    def register(self, expiration_in_seconds=360, re_register_time=180):
        """ Convenience function to register a SipEndpoint """
        if not expiration_in_seconds:
            return self.unregister()
        if re_register_time and self.registered:
            self.re_register_timer = Timer(re_register_time, self.register, (expiration_in_seconds, re_register_time))
            self.re_register_timer.start()
        flow.register(self, expiration_in_seconds)
        self.registered = True

    def unregister(self):
        """ Convenience function to un-register a SipEndpoint"""
        if self.re_register_timer:
            self.re_register_timer.cancel()
        flow.unregister(self)
        self.registered = False

    def free_resources(self, message):
        """ Free memory allocated to a cleared call """

        status_or_method = message.get_status_or_method()
        cseq_method = message["CSeq"].split()[1]
        call_ended_successfully = status_or_method.startswith("2") and cseq_method in ("BYE", "CANCEL")
        call_rejected_with_error = status_or_method[0] in "3456"
        if call_ended_successfully or call_rejected_with_error:
            dialog = message.get_dialog()
            with self.lock:
                dialog_index = self.dialogs.index(dialog)
                # self.dialogs.pop(dialog_index)
                # self.requests.pop(dialog_index)
                # self.tags.pop(dialog_hash(dialog))
                # self.last_messages_per_dialog.pop(dialog_index)
                # print("Cleared call number", dialog_index, "after message", status_or_method, "with cseq", cseq_method,
                #       ". Sizes are", len(self.dialogs), len(self.requests), len(self.tags))



"""\
Purpose: Define reusable SIP Flows
Initial Version: Costas Skarakis 11/11/2018
"""
from sip.messages import message
from time import sleep


def register(sip_ep, expiration_in_seconds=360):
    """ Register a SIP endpoint """
    sip_ep.parameters["expires"] = expiration_in_seconds
    sip_ep.send_new(message_string=message["Register_1"], expected_response="200 OK")


def unregister(sip_ep):
    """ Un-Register a SIP endpoint """
    register(sip_ep, expiration_in_seconds=0)


def basic_call(A, B, duration=2):
    """ Basic SIP Call flow between two SIP Endpoints
    A calls B. B answers. A hangs up after the duration.

    :A: The caller
    :B: The callee
    :duration: The duration of the call
    """
    A.send_new(B, message["Invite_SDP_1"], expected_response="Trying")
    B.waitForMessage("INVITE")
    B.reply(message["Trying_1"])
    B.reply(message["Ringing_1"])
    A.waitForMessage("180 Ringing")
    B.reply(message["200_OK_SDP_1"])
    B.waitForMessage("ACK")
    A.waitForMessage("200 OK")
    A.send(message["Ack_1"])

    sleep(duration)

    A.send(message["Bye_1"], expected_response="200 OK")
    B.waitForMessage("BYE")
    B.reply(message["200_OK_1"])